# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import collections
import logging
import weakref

from aiophoenixdb.avatica.proto.responses_pb import ResultSetResponse
from aiophoenixdb.avatica.proto import common_pb
from aiophoenixdb.errors import InternalError, ProgrammingError
from aiophoenixdb.types import TypeHelper

__all__ = ['Cursor', 'ColumnDescription', 'DictCursor']

logger = logging.getLogger(__name__)

# TODO see note in Cursor.rowcount()
MAX_INT = 2 ** 64 - 1

ColumnDescription = collections.namedtuple('ColumnDescription',
                                           ['name',
                                            'type_code',
                                            'display_size',
                                            'internal_size',
                                            'precision',
                                            'scale',
                                            'null_ok']
                                           )
"""Named tuple for representing results from :attr:`Cursor.description`."""


class CursorRef:

    def __init__(self, o, callback=None):
        self._ref = weakref.ref(o, (lambda x: callback(self) if callback is not None else None))

    def __call__(self, *args, **kwargs):
        return self._ref()


class Cursor(object):
    """Database cursor for executing queries and iterating over results.

    You should not construct this object manually, use :meth:
    `Connection.cursor() <aiophoenixdb.connection.Connection.cursor>` instead.
    """

    _ARRAY_SIZE = 1
    """
    Read/write attribute specifying the number of rows to fetch
    at a time with :meth:`fetchmany`. It defaults to 1 meaning to
    fetch a single row at a time.
    """

    _ITER_SIZE = 2000
    """
    Read/write attribute specifying the number of rows to fetch
    from the backend at each network roundtrip during iteration
    on the cursor. The default is 2000.
    """

    def __init__(self, connection, _id=-1):
        self._connection = connection
        self._id = _id
        self._signature = None
        self._column_data_types = []
        self._frame = None
        self._pos = 0
        self._closed = False
        self._array_size = self.__class__._ARRAY_SIZE
        self._iter_size = self.__class__._ITER_SIZE
        self._update_count = -1

    def __del__(self):
        if not self._connection.closed and not self._closed:
            asyncio.run(self.close())

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self._closed:
            await self.close()

    async def __aiter__(self):
        return self

    async def __anext__(self):
        row = await self.fetchone()
        if row is None:
            raise StopIteration
        return row

    async def close(self):
        """Closes the cursor.
        No further operations are allowed once the cursor is closed.

        If the cursor is used in a ``with`` statement, this method will
        be automatically called at the end of the ``with`` block.
        """
        if self._closed:
            raise ProgrammingError('The cursor is already closed.')
        if self._id is not None:
            await self._connection.client.close_statement(self._connection.connect_id, self._id)
            self._id = -1
        self._signature = None
        self._column_data_types = []
        self._frame = None
        self._pos = 0
        self._closed = True

    @property
    def closed(self):
        """Read-only attribute specifying if the cursor is closed or not."""
        return self._closed

    @property
    def description(self):
        if self._signature is None:
            return None
        description = []
        for column in self._signature.columns:
            description.append(ColumnDescription(
                self._get_column_name(column),
                column.type.name,
                column.display_size,
                None,
                column.precision,
                column.scale,
                None if column.nullable == 2 else bool(column.nullable),
            ))
        return description

    @staticmethod
    def _get_column_name(column):
        if column.label:
            # Not empty
            return column.label
        else:
            return column.column_name

    async def _set_id(self, _id):
        if self._id is not None and self._id != _id:
            await self._connection.client.close_statement(self._connection.connect_id, self._id)
        self._id = _id

    def _set_signature(self, signature):
        self._signature = signature
        self._column_data_types = []
        self._parameter_data_types = []
        if signature is None:
            return

        for column in signature.columns:
            dtype = TypeHelper.from_column(column)
            self._column_data_types.append(dtype)

        for parameter in signature.parameters:
            dtype = TypeHelper.from_param(parameter)
            self._parameter_data_types.append(dtype)

    def _set_frame(self, frame):
        self._frame = frame
        self._pos = 0

        if frame is not None:
            if frame.rows:
                self._pos = 0
            elif not frame.done:
                raise InternalError('Got an empty frame, but the statement is not done yet.')

    async def _fetch_next_frame(self):
        offset = self._frame.offset + len(self._frame.rows)
        frame = await self._connection.client.fetch(
            self._connection.connect_id, self._id,
            offset=offset, frame_max_size=self._iter_size)
        self._set_frame(frame)

    async def process_result(self, result: ResultSetResponse):
        if result.own_statement:
            await self._set_id(result.statement_id)
        self._set_signature(result.signature)
        self._set_frame(result.first_frame)
        self._update_count = result.update_count

    async def _process_results(self, results):
        if results:
            await self.process_result(results[0])

    def _transform_parameters(self, parameters):
        if len(parameters) != len(self._parameter_data_types):
            raise ProgrammingError('Number of placeholders (?) must match number of parameters.'
                                   ' Number of placeholders: {0}. Number of parameters: {1}'
                                   .format(len(self._parameter_data_types), len(parameters)))
        typed_parameters = []
        for value, data_type in zip(parameters, self._parameter_data_types):
            field_name, rep, mutate_to, cast_from, is_array = data_type
            typed_value = common_pb.TypedValue()

            if value is None:
                typed_value.null = True
                typed_value.type = common_pb.Rep.NULL
            else:
                typed_value.null = False
                if is_array:
                    if type(value) in [list, tuple]:
                        for element in value:
                            if mutate_to is not None:
                                element = mutate_to(element)
                            typed_element = common_pb.TypedValue()
                            if element is None:
                                typed_element.null = True
                            else:
                                typed_element.type = rep
                                setattr(typed_element, field_name, element)
                            typed_value.array_value.append(typed_element)
                        typed_value.type = common_pb.Rep.ARRAY
                        typed_value.component_type = rep
                    else:
                        raise ProgrammingError('Scalar value specified for array parameter.')
                else:
                    if mutate_to is not None:
                        value = mutate_to(value)
                    typed_value.type = rep
                    setattr(typed_value, field_name, value)

            typed_parameters.append(typed_value)
        return typed_parameters

    async def execute(self, operation, parameters=None):
        if self._closed:
            raise ProgrammingError('The cursor is already closed.')
        self._update_count = -1
        self._set_frame(None)
        if parameters is None:
            if self._id is None:
                c_id = await self._connection.client.create_statement(self._connection.connect_id)
                await self._set_id(c_id)
            results = await self._connection.client.prepare_and_execute(
                self._connection.connect_id, self._id,
                operation, first_frame_max_size=self._iter_size)
            await self._process_results(results)
        else:
            statement = await self._connection.client.prepare(
                self._connection.connect_id, operation)
            await self._set_id(statement.id)
            self._set_signature(statement.signature)

            results = await self._connection.client.execute(
                self._connection.connect_id, self._id,
                statement.signature, self._transform_parameters(parameters),
                first_frame_max_size=self._iter_size)
            await self._process_results(results)

    async def executemany(self, operation, seq_of_parameters):
        if self._closed:
            raise ProgrammingError('The cursor is already closed.')
        self._update_count = -1
        self._set_frame(None)
        statement = await self._connection.client.prepare(
            self._connection.connect_id, operation, max_rows_total=0)
        await self._set_id(statement.id)
        self._set_signature(statement.signature)
        return await self._connection.client.execute_batch(
            self._connection.connect_id, self._id,
            [self._transform_parameters(p) for p in seq_of_parameters])

    async def get_sync_results(self, state):
        if self._closed:
            raise ProgrammingError('The cursor is already closed.')
        if self._id is None:
            c_id = await self._connection.client.create_statement(self._connection.connect_id)
            await self._set_id(c_id)
        return self._connection.client.get_sync_results(self._connection.connect_id, self._id, state)

    async def fetch(self, signature):
        if self._closed:
            raise ProgrammingError('The cursor is already closed.')
        self._update_count = -1
        self._set_signature(signature)
        frame = await self._connection.client.fetch(self._connection.connect_id, self._id, 0, self._iter_size)
        self._set_frame(frame)

    def transform_row(self, row):
        """Transforms a Row into Python values.

        :param row:
            A ``common_pb2.Row`` object.

        :returns:
            A list of values cast into the correct Python types.

        :raises:
            NotImplementedError
        """
        tmp_row = []

        for i, column in enumerate(row.value):
            if column.scalar_value.null:
                tmp_row.append(None)
            elif column.has_array_value:
                field_name, rep, mutate_to, cast_from = self._column_data_types[i]

                list_value = []
                for j, typed_value in enumerate(column.array_value):
                    value = getattr(typed_value, field_name)
                    if cast_from is not None:
                        value = cast_from(value)
                    list_value.append(value)

                tmp_row.append(list_value)
            else:
                field_name, rep, mutate_to, cast_from = self._column_data_types[i]

                # get the value from the field_name
                value = getattr(column.scalar_value, field_name)

                # cast the value
                if cast_from is not None:
                    value = cast_from(value)

                tmp_row.append(value)
        return tmp_row

    async def fetchone(self):
        if self._frame is None:
            raise ProgrammingError('No select statement was executed.')
        if self._pos is None:
            return None
        rows = self._frame.rows
        row = self.transform_row(rows[self._pos])
        self._pos += 1
        if self._pos >= len(rows):
            self._pos = 0
            if not self._frame.done:
                await self._fetch_next_frame()
        return row

    async def fetchmany(self, size=None):
        if size is None:
            size = self._array_size
        rows = []
        while size > 0:
            row = await self.fetchone()
            if row is None:
                break
            rows.append(row)
            size -= 1
        return rows

    async def fetchall(self):
        rows = []
        while True:
            row = await self.fetchone()
            if row is None:
                break
            rows.append(row)
        return rows

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size, column=None):
        pass

    @property
    def connection(self):
        """Read-only attribute providing access to the :class:`Connection <aiophoenixdb.connection.Connection>`
        object this cursor was created from."""
        return self._connection

    @property
    def rowcount(self):
        """Read-only attribute specifying the number of rows affected by
        the last executed DML statement or -1 if the number cannot be
        determined. Note that this will always be set to -1 for select
        queries."""
        # TODO instead of -1, this ends up being set to Integer.MAX_VALUE
        if self._update_count == MAX_INT:
            return -1
        return self._update_count

    @property
    def rownumber(self):
        """Read-only attribute providing the current 0-based index of the
        cursor in the result set or ``None`` if the index cannot be
        determined.

        The index can be seen as index of the cursor in a sequence
        (the result set). The next fetch operation will fetch the
        row indexed by :attr:`rownumber` in that sequence.
        """
        if self._frame is not None and self._pos is not None:
            return self._frame.offset + self._pos
        return self._pos

    def ref(self, callback=None):
        return CursorRef(self, callback)


class DictCursor(Cursor):
    """A cursor which returns results as a dictionary"""

    def transform_row(self, row):
        row = super(__class__).transform_row(row)
        d = {}
        for ind, val in enumerate(row):
            d[self._get_column_name(self._signature.columns[ind])] = val
        return d



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import betterproto


class StatementType(betterproto.Enum):
    """Has to be consistent with Meta.StatementType"""

    SELECT: int
    INSERT: int
    UPDATE: int
    DELETE: int
    UPSERT: int
    MERGE: int
    OTHER_DML: int
    CREATE: int
    DROP: int
    ALTER: int
    OTHER_DDL: int
    CALL: int


class Rep(betterproto.Enum):
    PRIMITIVE_BOOLEAN: int
    PRIMITIVE_BYTE: int
    PRIMITIVE_CHAR: int
    PRIMITIVE_SHORT: int
    PRIMITIVE_INT: int
    PRIMITIVE_LONG: int
    PRIMITIVE_FLOAT: int
    PRIMITIVE_DOUBLE: int
    BOOLEAN: int
    BYTE: int
    CHARACTER: int
    SHORT: int
    INTEGER: int
    LONG: int
    FLOAT: int
    DOUBLE: int
    BIG_INTEGER: int
    BIG_DECIMAL: int
    JAVA_SQL_TIME: int
    JAVA_SQL_TIMESTAMP: int
    JAVA_SQL_DATE: int
    JAVA_UTIL_DATE: int
    BYTE_STRING: int
    STRING: int
    NUMBER: int
    OBJECT: int
    NULL: int
    ARRAY: int
    STRUCT: int
    MULTISET: int


class Severity(betterproto.Enum):
    """
    The severity of some unexpected outcome to an operation. Protobuf enum
    values must be unique across all other enums
    """

    UNKNOWN_SEVERITY: int
    FATAL_SEVERITY: int
    ERROR_SEVERITY: int
    WARNING_SEVERITY: int


class MetaDataOperation(betterproto.Enum):
    """Enumeration corresponding to DatabaseMetaData operations"""

    GET_ATTRIBUTES: int
    GET_BEST_ROW_IDENTIFIER: int
    GET_CATALOGS: int
    GET_CLIENT_INFO_PROPERTIES: int
    GET_COLUMN_PRIVILEGES: int
    GET_COLUMNS: int
    GET_CROSS_REFERENCE: int
    GET_EXPORTED_KEYS: int
    GET_FUNCTION_COLUMNS: int
    GET_FUNCTIONS: int
    GET_IMPORTED_KEYS: int
    GET_INDEX_INFO: int
    GET_PRIMARY_KEYS: int
    GET_PROCEDURE_COLUMNS: int
    GET_PROCEDURES: int
    GET_PSEUDO_COLUMNS: int
    GET_SCHEMAS: int
    GET_SCHEMAS_WITH_ARGS: int
    GET_SUPER_TABLES: int
    GET_SUPER_TYPES: int
    GET_TABLE_PRIVILEGES: int
    GET_TABLES: int
    GET_TABLE_TYPES: int
    GET_TYPE_INFO: int
    GET_UDTS: int
    GET_VERSION_COLUMNS: int


class StateType(betterproto.Enum):
    SQL: int
    METADATA: int


class CursorFactoryStyle(betterproto.Enum):
    OBJECT: int
    RECORD: int
    RECORD_PROJECTION: int
    ARRAY: int
    LIST: int
    MAP: int


class MetaDataOperationArgumentArgumentType(betterproto.Enum):
    STRING: int
    BOOL: int
    INT: int
    REPEATED_STRING: int
    REPEATED_INT: int
    NULL: int


class ConnectionProperties(betterproto.Message):
    """Details about a connection"""

    is_dirty: bool
    auto_commit: bool
    has_auto_commit: bool
    read_only: bool
    has_read_only: bool
    transaction_isolation: int
    catalog: str
    schema: str


class StatementHandle(betterproto.Message):
    """Statement handle"""

    connection_id: str
    id: int
    signature: "Signature"


class Signature(betterproto.Message):
    """Results of preparing a statement"""

    columns: List["ColumnMetaData"]
    sql: str
    parameters: List["AvaticaParameter"]
    cursor_factory: "CursorFactory"
    statement_type: "StatementType"


class ColumnMetaData(betterproto.Message):
    ordinal: int
    auto_increment: bool
    case_sensitive: bool
    searchable: bool
    currency: bool
    nullable: int
    signed: bool
    display_size: int
    label: str
    column_name: str
    schema_name: str
    precision: int
    scale: int
    table_name: str
    catalog_name: str
    read_only: bool
    writable: bool
    definitely_writable: bool
    column_class_name: str
    type: "AvaticaType"


class AvaticaType(betterproto.Message):
    """Base class for a column type"""

    id: int
    name: str
    rep: "Rep"
    columns: List["ColumnMetaData"]
    component: "AvaticaType"


class AvaticaParameter(betterproto.Message):
    """Metadata for a parameter"""

    signed: bool
    precision: int
    scale: int
    parameter_type: int
    type_name: str
    class_name: str
    name: str


class CursorFactory(betterproto.Message):
    """Information necessary to convert an Iterable into a Calcite Cursor"""

    style: "CursorFactoryStyle"
    class_name: str
    field_names: List[str]


class Frame(betterproto.Message):
    """A collection of rows"""

    offset: int
    done: bool
    rows: List["Row"]


class Row(betterproto.Message):
    """A row is a collection of values"""

    value: List["ColumnValue"]


class DatabaseProperty(betterproto.Message):
    """
    Database property, list of functions the database provides for a certain
    operation
    """

    name: str
    functions: List[str]


class WireMessage(betterproto.Message):
    """
    Message which encapsulates another message to support a single RPC endpoint
    """

    name: str
    wrapped_message: bytes


class ColumnValue(betterproto.Message):
    """A value might be a TypedValue or an Array of TypedValue's"""

    value: List["TypedValue"]
    array_value: List["TypedValue"]
    has_array_value: bool
    scalar_value: "TypedValue"


class TypedValue(betterproto.Message):
    """
    Generic wrapper to support any SQL type. Struct-like to work around no
    polymorphism construct.
    """

    type: "Rep"
    bool_value: bool
    string_value: str
    number_value: int
    # includes numeric types and date/time types.
    bytes_value: bytes
    double_value: float
    null: bool
    array_value: List["TypedValue"]
    component_type: "Rep"
    implicitly_null: bool


class MetaDataOperationArgument(betterproto.Message):
    """Represents the breadth of arguments to DatabaseMetaData functions"""

    string_value: str
    bool_value: bool
    int_value: int
    string_array_values: List[str]
    int_array_values: List[int]
    type: "MetaDataOperationArgumentArgumentType"


class QueryState(betterproto.Message):
    type: "StateType"
    sql: str
    op: "MetaDataOperation"
    args: List["MetaDataOperationArgument"]
    has_args: bool
    has_sql: bool
    has_op: bool



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gssapi

from aiophoenixdb import errors, types
from aiophoenixdb.avatica.client import AvaticaClient
from aiophoenixdb.connection import Connection
from aiophoenixdb.errors import *  # noqa: F401,F403
from aiophoenixdb.types import *  # noqa: F401,F403

from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from requests_gssapi import HTTPSPNEGOAuth
from urllib.parse import urlencode, urlparse, urlunparse, parse_qs

__all__ = ['connect', 'apilevel', 'threadsafety', 'paramstyle'] + types.__all__ + errors.__all__

apilevel = "2.0"
"""
This module supports the `DB API 2.0 interface <https://www.python.org/dev/peps/pep-0249/>`_.
"""

threadsafety = 1
"""
Multiple threads can share the module, but neither connections nor cursors.
"""

paramstyle = 'qmark'
"""
Parmetrized queries should use the question mark as a parameter placeholder.

For example::

 cursor.execute("SELECT * FROM table WHERE id = ?", [my_id])
"""


async def connect(url, max_retries=None, auth=None, authentication=None, avatica_user=None, avatica_password=None,
                  truststore=None, verify=None, do_as=None, user=None, password=None, extra_headers=None, **kwargs):
    """Connects to a Phoenix query server.

    :param url:
        URL to the Phoenix query server, e.g. ``http://localhost:8765/``

    :param autocommit:
        Switch the connection to autocommit mode.

    :param readonly:
        Switch the connection to readonly mode.

    :param max_retries:
        The maximum number of retries in case there is a connection error.

    :param cursor_factory:
        If specified, the connection's :attr:`~aiophoenixdb.connection.Connection.cursor_factory`
        is set to it.

    :param auth:
        Authentication configuration object as expected by the underlying python_requests and
        python_requests_gssapi library

    :param verify:
        The path to the PEM file for verifying the server's certificate. It is passed directly to
        the `~verify` parameter of the underlying python_requests library.
        Setting it to False disables the server certificate verification.

    :param do_as:
        Username to impersonate (sets the Hadoop doAs URL parameter)

    :param authentication:
        Alternative way to specify the authentication mechanism that mimics
        the semantics of the JDBC drirver

    :param avatica_user:
        Username for BASIC or DIGEST authentication. Use in conjunction with the
        `~authentication' option.

    :param avatica_password:
        Password for BASIC or DIGEST authentication. Use in conjunction with the
        `~authentication' option.

    param user
        If `~authentication' is BASIC or DIGEST then alias for `~avatica_user`
        If `~authentication' is NONE or SPNEGO then alias for `~do_as`

    param password
        If `~authentication` is BASIC or DIGEST then is alias for `~avatica_password`

    :param truststore:
        Alias for verify

    :param extra_headers:
        Additional HTTP headers as a dictionary

    :returns:
        :class:`~aiophoenixdb.connection.Connection` object.
    """

    (url, auth, verify) = _process_args(
        url, auth=auth, authentication=authentication,
        avatica_user=avatica_user, avatica_password=avatica_password,
        truststore=truststore, verify=verify, do_as=do_as, user=user, password=password)

    client = AvaticaClient(url, max_retries=max_retries, auth=auth, verify=verify, extra_headers=extra_headers)
    conn = Connection(client, **kwargs)
    await conn.connect()
    return conn


def _get_SPNEGOAuth():
    try:
        spnego = gssapi.mechs.Mechanism.from_sasl_name("SPNEGO")
    except AttributeError:
        spnego = gssapi.OID.from_int_seq("1.3.6.1.5.5.2")
    return HTTPSPNEGOAuth(opportunistic_auth=True, mech=spnego)


def _process_args(
        url, auth=None, authentication=None, avatica_user=None, avatica_password=None,
        truststore=None, verify=None, do_as=None, user=None, password=None):
    url_parsed = urlparse(url)
    url_params = parse_qs(url_parsed.query, keep_blank_values=True)

    # Parse supported JDBC compatible parameters from URL. args have precendece
    # Unlike the JDBC driver, we are expecting these as query params, as the avatica java client
    # has a different idea of what a URL param is than urlparse. (urlparse seems just broken
    # in this regard)
    params_changed = False
    if auth is None and authentication is None and 'authentication' in url_params:
        authentication = url_params['authentication'][0]
        del url_params['authentication']
        params_changed = True

    if avatica_user is None and 'avatica_user' in url_params:
        avatica_user = url_params['avatica_user'][0]
        del url_params['avatica_user']
        params_changed = True

    if avatica_password is None and 'avatica_password' in url_params:
        avatica_password = url_params['avatica_password'][0]
        del url_params['avatica_password']
        params_changed = True

    if verify is None and truststore is None and 'truststore' in url_params:
        truststore = url_params['truststore'][0]
        del url_params['truststore']
        params_changed = True

    if authentication == 'BASIC' or authentication == 'DIGEST':
        # Handle standard user and password parameters
        if user is not None and avatica_user is None:
            avatica_user = user
        if password is not None and avatica_password is None:
            avatica_password = password
    else:
        # interpret standard user parameter as do_as for SPNEGO and NONE
        if user is not None and do_as is None:
            do_as = user

    # Add doAs
    if do_as:
        url_params['doAs'] = do_as
        params_changed = True

    if params_changed:
        url_parsed = url_parsed._replace(query=urlencode(url_params))
        url = urlunparse(url_parsed)

    if auth == "SPNEGO":
        # Special case for backwards compatibility
        auth = _get_SPNEGOAuth()
    elif auth is None and authentication is not None:
        if authentication == "SPNEGO":
            auth = _get_SPNEGOAuth()
        elif authentication == "BASIC" and avatica_user is not None and avatica_password is not None:
            auth = HTTPBasicAuth(avatica_user, avatica_password)
        elif authentication == "DIGEST" and avatica_user is not None and avatica_password is not None:
            auth = HTTPDigestAuth(avatica_user, avatica_password)

    if verify is None and truststore is not None:
        verify = truststore

    return url, auth, verify



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import sys
import time
from decimal import Decimal
from typing import Dict
from aiophoenixdb.avatica.proto import common_pb


__all__ = [
    'to_date', 'to_time', 'to_timestamp', 'to_date_from_ticks', 'to_time_from_ticks', 'to_timestamp_from_ticks',
    'to_binary', 'STRING', 'BINARY', 'NUMBER', 'DATETIME', 'ROWID', 'BOOLEAN',
    'TypeHelper',
]


def to_date(year, month, day):
    """Constructs an object holding a date value."""
    return datetime.date(year, month, day)


def to_time(hour, minute, second):
    """Constructs an object holding a time value."""
    return datetime.time(hour, minute, second)


def to_timestamp(year, month, day, hour, minute, second):
    """Constructs an object holding a datetime/timestamp value."""
    return datetime.datetime(year, month, day, hour, minute, second)


def to_date_from_ticks(ticks):
    """Constructs an object holding a date value from the given UNIX timestamp."""
    return to_date(*time.localtime(ticks)[:3])


def to_time_from_ticks(ticks):
    """Constructs an object holding a time value from the given UNIX timestamp."""
    return to_time(*time.localtime(ticks)[3:6])


def to_timestamp_from_ticks(ticks):
    """Constructs an object holding a datetime/timestamp value from the given UNIX timestamp."""
    return to_timestamp(*time.localtime(ticks)[:6])


def to_binary(value):
    """Constructs an object capable of holding a binary (long) string value."""
    return bytes(value)


def time_from_java_sql_time(n):
    dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=n)
    return dt.time()


def time_to_java_sql_time(t):
    return ((t.hour * 60 + t.minute) * 60 + t.second) * 1000 + t.microsecond // 1000


def date_from_java_sql_date(n):
    return datetime.date(1970, 1, 1) + datetime.timedelta(days=n)


def date_to_java_sql_date(d):
    if isinstance(d, datetime.datetime):
        d = d.date()
    td = d - datetime.date(1970, 1, 1)
    return td.days


def datetime_from_java_sql_timestamp(n):
    return datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=n)


def datetime_to_java_sql_timestamp(d):
    td = d - datetime.datetime(1970, 1, 1)
    return td.microseconds // 1000 + (td.seconds + td.days * 24 * 3600) * 1000


# FIXME This doesn't seem to be used anywhere in the code
class ColumnType(object):

    def __init__(self, eq_types):
        self.eq_types = tuple(eq_types)
        self.eq_types_set = set(eq_types)

    def __eq__(self, other):
        return other in self.eq_types_set

    def __cmp__(self, other):
        if other in self.eq_types_set:
            return 0
        if other < self.eq_types:
            return 1
        else:
            return -1


STRING = ColumnType(['VARCHAR', 'CHAR'])
"""Type object that can be used to describe string-based columns."""

BINARY = ColumnType(['BINARY', 'VARBINARY'])
"""Type object that can be used to describe (long) binary columns."""

NUMBER = ColumnType([
    'INTEGER', 'UNSIGNED_INT', 'BIGINT', 'UNSIGNED_LONG', 'TINYINT', 'UNSIGNED_TINYINT',
    'SMALLINT', 'UNSIGNED_SMALLINT', 'FLOAT', 'UNSIGNED_FLOAT', 'DOUBLE', 'UNSIGNED_DOUBLE', 'DECIMAL'
])
"""Type object that can be used to describe numeric columns."""

DATETIME = ColumnType(['TIME', 'DATE', 'TIMESTAMP', 'UNSIGNED_TIME', 'UNSIGNED_DATE', 'UNSIGNED_TIMESTAMP'])
"""Type object that can be used to describe date/time columns."""

ROWID = ColumnType([])
"""Only implemented for DB API 2.0 compatibility, not used."""

BOOLEAN = ColumnType(['BOOLEAN'])
"""Type object that can be used to describe boolean columns. This is a phoenixdb-specific extension."""

if sys.version_info[0] < 3:
    _long = long  # noqa: F821
else:
    _long = int

FIELD_MAP = {
    'bool_value': [
        (common_pb.Rep.BOOLEAN, None, None),
        (common_pb.Rep.PRIMITIVE_BOOLEAN, None, None),
    ],
    'string_value': [
        (common_pb.Rep.CHARACTER, None, None),
        (common_pb.Rep.PRIMITIVE_CHAR, None, None),
        (common_pb.Rep.STRING, None, None),
        (common_pb.Rep.BIG_DECIMAL, str, Decimal),
    ],
    'number_value': [
        (common_pb.Rep.INTEGER, None, int),
        (common_pb.Rep.PRIMITIVE_INT, None, int),
        (common_pb.Rep.SHORT, None, int),
        (common_pb.Rep.PRIMITIVE_SHORT, None, int),
        (common_pb.Rep.LONG, None, _long),
        (common_pb.Rep.PRIMITIVE_LONG, None, _long),
        (common_pb.Rep.BYTE, None, int),
        (common_pb.Rep.JAVA_SQL_TIME, time_to_java_sql_time, time_from_java_sql_time),
        (common_pb.Rep.JAVA_SQL_DATE, date_to_java_sql_date, date_from_java_sql_date),
        (common_pb.Rep.JAVA_SQL_TIMESTAMP, datetime_to_java_sql_timestamp, datetime_from_java_sql_timestamp),
    ],
    'bytes_value': [
        (common_pb.Rep.BYTE_STRING, to_binary, None),
    ],
    'double_value': [
        (common_pb.Rep.DOUBLE, float, float),
        (common_pb.Rep.PRIMITIVE_DOUBLE, float, float)
    ]
}
"""The master map that describes how to handle types, keyed by TypedData field"""

REP_MAP = dict((v[0], (k, v[0], v[1], v[2])) for k in FIELD_MAP for v in FIELD_MAP[k])
"""Flips the available types to allow for faster lookup by protobuf Rep

This mapping should be structured as:
    {
        'common_pb2.BIG_DECIMAL': ('string_value', common_pb2.BIG_DECIMAL, str, Decimal),),
        ...
        '<Rep enum>': (<field_name>, <mutate_to function>, <cast_from function>),
    }
"""

JDBC_TO_REP: Dict[int, common_pb.Rep] = dict([
    # These are the standard types that are used in Phoenix
    (-6, common_pb.Rep.BYTE),  # TINYINT
    (5, common_pb.Rep.SHORT),  # SMALLINT
    (4, common_pb.Rep.INTEGER),  # INTEGER
    (-5, common_pb.Rep.LONG),  # BIGINT
    (6, common_pb.Rep.DOUBLE),  # FLOAT
    (8, common_pb.Rep.DOUBLE),  # DOUBLE
    (2, common_pb.Rep.BIG_DECIMAL),  # NUMERIC
    (1, common_pb.Rep.STRING),  # CHAR
    (91, common_pb.Rep.JAVA_SQL_DATE),  # DATE
    (92, common_pb.Rep.JAVA_SQL_TIME),  # TIME
    (93, common_pb.Rep.JAVA_SQL_TIMESTAMP),  # TIMESTAMP
    (-2, common_pb.Rep.BYTE_STRING),  # BINARY
    (-3, common_pb.Rep.BYTE_STRING),  # VARBINARY
    (16, common_pb.Rep.BOOLEAN),  # BOOLEAN
    # These are the Non-standard types defined by Phoenix
    (19, common_pb.Rep.JAVA_SQL_DATE),  # UNSIGNED_DATE
    (15, common_pb.Rep.DOUBLE),  # UNSIGNED_DOUBLE
    (14, common_pb.Rep.DOUBLE),  # UNSIGNED_FLOAT
    (9, common_pb.Rep.INTEGER),  # UNSIGNED_INT
    (10, common_pb.Rep.LONG),  # UNSIGNED_LONG
    (13, common_pb.Rep.SHORT),  # UNSIGNED_SMALLINT
    (20, common_pb.Rep.JAVA_SQL_TIMESTAMP),  # UNSIGNED_TIMESTAMP
    (11, common_pb.Rep.BYTE),  # UNSIGNED_TINYINT
    # The following are not used by Phoenix, but some of these are used by Avaticafor
    # parameter types
    (-7, common_pb.Rep.BOOLEAN),  # BIT
    (7, common_pb.Rep.DOUBLE),  # REAL
    (3, common_pb.Rep.BIG_DECIMAL),  # DECIMAL
    (12, common_pb.Rep.STRING),  # VARCHAR
    (-1, common_pb.Rep.STRING),  # LONGVARCHAR
    (-4, common_pb.Rep.BYTE_STRING),  # LONGVARBINARY
    (2004, common_pb.Rep.BYTE_STRING),  # BLOB
    (2005, common_pb.Rep.STRING),  # CLOB
    (-15, common_pb.Rep.STRING),  # NCHAR
    (-9, common_pb.Rep.STRING),  # NVARCHAR
    (-16, common_pb.Rep.STRING),  # LONGNVARCHAR
    (2011, common_pb.Rep.STRING),  # NCLOB
    (2009, common_pb.Rep.STRING),  # SQLXML
    # Returned by Avatica for Arrays in EMPTY resultsets
    (2000, common_pb.Rep.BYTE_STRING)  # JAVA_OBJECT
    # These are defined by JDBC, but cannot be mapped
    # NULL
    # OTHER
    # DISTINCT
    # STRUCT
    # ARRAY 2003 - We are handling this as a special case
    # REF
    # DATALINK
    # ROWID
    # REF_CURSOR
    # TIME WITH TIMEZONE
    # TIMESTAMP WITH TIMEZONE

    ])
"""Maps the JDBC Type IDs to Protobuf Reps """

JDBC_MAP = {}
for k, v in JDBC_TO_REP.items():
    JDBC_MAP[k & 0xffffffff] = REP_MAP[v]
"""Flips the available types to allow for faster lookup by JDBC type ID

It has the same format as REP_MAP, but is keyed by JDBC type ID
"""


class TypeHelper(object):

    @staticmethod
    def from_param(param):
        """Retrieves a field name and functions to cast to/from based on an AvaticaParameter object

        :param param:
            Protobuf AvaticaParameter object

        :returns: tuple ``(field_name, rep, mutate_to, cast_from, is_array)``
            WHERE
            ``field_name`` is the attribute in ``common_pb2.TypedValue``
            ``rep`` is the common_pb2.Rep enum
            ``mutate_to`` is the function to cast values into Phoenix values, if any
            ``cast_from`` is the function to cast from the Phoenix value to the Python value, if any
            ``is_array`` the param expects an array instead of scalar

        :raises:
            NotImplementedError
        """
        jdbc_code = param.parameter_type
        if 2900 < jdbc_code < 3100:
            return TypeHelper._from_jdbc(jdbc_code-3000) + (True,)
        else:
            return TypeHelper._from_jdbc(jdbc_code) + (False,)

    @staticmethod
    def from_column(column):
        """Retrieves a field name and functions to cast to/from based on a TypedValue object

        :param column:
            Protobuf TypedValue object

        :returns: tuple ``(field_name, rep, mutate_to, cast_from)``
            WHERE
            ``field_name`` is the attribute in ``common_pb2.TypedValue``
            ``rep`` is the common_pb2.Rep enum
            ``mutate_to`` is the function to cast values into Phoenix values, if any
            ``cast_from`` is the function to cast from the Phoenix value to the Python value, if any

        :raises:
            NotImplementedError
        """
        if column.type.id == 2003:
            return TypeHelper._from_jdbc(column.type.component.id)
        else:
            return TypeHelper._from_jdbc(column.type.id)

    @staticmethod
    def _from_jdbc(jdbc_code):
        if jdbc_code not in JDBC_MAP:
            # This should not happen. It's either a bug, or Avatica has added new types
            raise NotImplementedError('JDBC TYPE CODE {} is not supported'.format(jdbc_code))

        return JDBC_MAP[jdbc_code]



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import sys
from typing import Dict, Any, List, Tuple, Set

from aiophoenixdb.avatica.proto import common_pb

__all__: List[str]

def to_date(year: int, month: int, day: int) -> datetime.date: ...


def to_time(hour: int, minute: int, second: int) -> datetime.time: ...


def to_timestamp(year: int, month: int, day: int, hour: int, minute: int, second: int) -> datetime.datetime: ...


def to_date_from_ticks(ticks) -> datetime.date: ...


def to_time_from_ticks(ticks) -> datetime.time: ...


def to_timestamp_from_ticks(ticks) -> datetime.datetime: ...


def to_binary(value) -> bytes: ...


def time_from_java_sql_time(n) -> datetime.time: ...


def time_to_java_sql_time(t) -> int: ...


def date_from_java_sql_date(n) -> datetime.date: ...


def date_to_java_sql_date(d) -> int: ...


def datetime_from_java_sql_timestamp(n) -> datetime: ...


def datetime_to_java_sql_timestamp(d) -> int: ...


# FIXME This doesn't seem to be used anywhere in the code
class ColumnType(object):

    eq_types: Tuple[str]
    eq_types_set: Set[str]

    def __init__(self, eq_types: List[str]): ...

    def __eq__(self, other) -> bool: ...

    def __cmp__(self, other) -> int: ...


STRING: ColumnType

BINARY: ColumnType

NUMBER: ColumnType

DATETIME: ColumnType

ROWID: ColumnType

BOOLEAN : ColumnType

if sys.version_info[0] < 3:
    _long = long  # noqa: F821
else:
    _long = int

FIELD_MAP: Dict[str, Any]

REP_MAP: Dict[common_pb.Rep, Any]

JDBC_TO_REP: Dict[int, common_pb.Rep]

JDBC_MAP: Dict[int, Any]


class TypeHelper(object):

    @staticmethod
    def from_param(param) -> Any: ...

    @staticmethod
    def from_column(column) -> Any: ...

    @staticmethod
    def _from_jdbc(jdbc_code) -> Any: ...


# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aiohttp
import betterproto
from html.parser import HTMLParser
from typing import TypeVar, Type, List, Dict, Any, Optional
from urllib.parse import ParseResult
from aiohttp import ClientResponse, BasicAuth
from aiophoenixdb import errors
from aiophoenixdb.typeshed import Self
from .proto.common_pb import Frame
from .proto.common_pb import StatementHandle
from .proto.common_pb import ConnectionProperties
from .proto.responses_pb import ResultSetResponse
from .proto.responses_pb import SyncResultsResponse
from .proto.responses_pb import OpenConnectionResponse
from .proto.responses_pb import CloseConnectionResponse
from .proto.responses_pb import CloseStatementResponse
from .proto.responses_pb import CommitResponse
from .proto.responses_pb import RollbackResponse


class JettyErrorPageParser(HTMLParser):
    path: List[Any]
    title: List[Any]
    message: List[Any]

    def handle_starttag(self, tag, attrs) -> None: ...

    def handle_endtag(self, tag) -> None: ...

    def handle_data(self, data) -> None: ...


def parse_url(url) -> ParseResult:...


SQLSTATE_ERROR_CLASSES = [
    ('08', errors.OperationalError),  # Connection Exception
    ('22018', errors.IntegrityError),  # Constraint violatioin.
    ('22', errors.DataError),  # Data Exception
    ('23', errors.IntegrityError),  # Constraint Violation
    ('24', errors.InternalError),  # Invalid Cursor State
    ('25', errors.InternalError),  # Invalid Transaction State
    ('42', errors.ProgrammingError),  # Syntax Error or Access Rule Violation
    ('XLC', errors.OperationalError),  # Execution exceptions
    ('INT', errors.InternalError),  # Phoenix internal error
]

def raise_sql_error(code, sqlstate, message) -> None: ...


def parse_and_raise_sql_error(message: str) -> None: ...


def parse_error_page(html: str) -> None: ...


def parse_error_protobuf(text: bytes) -> None: ...


_MESSAGE_TYPE = TypeVar("_MESSAGE_TYPE", bound=betterproto.Message)


class AvaticaClient(object):
    _url: str
    _headers = {'content-type': 'application/x-google-protobuf'}
    _verify: Any
    _max_retries: int
    _session: aiohttp.ClientSession

    def __init__(self, url: str, max_retries: int, verify, extra_headers: Dict, auth: Optional[BasicAuth]): ...

    async def __post_request(self, body: bytes, retry_delay: int = 1) -> ClientResponse: ...

    async def _request(self, request_data: betterproto.Message,
                       expected_response_cls: Type[_MESSAGE_TYPE] | None = None) -> _MESSAGE_TYPE: ...

    async def get_catalogs(self, connection_id: str) -> ResultSetResponse: ...

    async def get_schemas(self, connection_id: str,
                          catalog: str | None = None,
                          schema_pattern: str | None = None) -> ResultSetResponse: ...

    async def get_tables(self, connection_id, catalog=None,
                         schema_pattern=None,
                         table_name_pattern=None,
                         type_list: List[str] = None) -> ResultSetResponse: ...

    async def get_columns(self, connection_id,
                          catalog=None, schema_pattern=None,
                          table_name_pattern=None,
                          column_name_pattern=None) -> ResultSetResponse: ...

    async def get_table_types(self, connection_id) -> ResultSetResponse: ...

    async def get_type_info(self, connection_id) -> ResultSetResponse: ...

    async def get_sync_results(self, connection_id, statement_id, state) -> SyncResultsResponse: ...

    async def connection_sync_dict(self, connection_id, conn_props: Dict[str, Any] = None) -> Dict[str, Any]: ...

    async def connection_sync(self, connection_id, conn_props: Dict[str, Any] = None) -> ConnectionProperties: ...

    async def open_connection(self, connection_id, info=None) -> OpenConnectionResponse: ...

    async def close_connection(self, connection_id) -> CloseConnectionResponse: ...

    async def create_statement(self, connection_id) -> int: ...

    async def close_statement(self, connection_id, statement_id) -> CloseStatementResponse: ...

    async def prepare_and_execute(self, connection_id, statement_id, sql, max_rows_total=None,
                                  first_frame_max_size=None) -> List[ResultSetResponse]: ...

    async def prepare(self, connection_id, sql, max_rows_total=None) -> StatementHandle: ...

    async def execute(self, connection_id,
                      statement_id,
                      signature,
                      parameter_values=None,
                      first_frame_max_size=None) -> List[ResultSetResponse]: ...

    async def execute_batch(self, connection_id, statement_id, rows) -> List[int]: ...

    async def fetch(self, connection_id, statement_id, offset=0, frame_max_size=None) -> Frame: ...

    async def commit(self, connection_id) -> CommitResponse: ...

    async def rollback(self, connection_id) -> RollbackResponse: ...

    async def close(self) -> None: ...

    async def __aenter__(self: Self) -> Self: ...

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None: ...


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: responses.proto
# plugin: python-betterproto
from .common_pb import *
from typing import List
import betterproto


@dataclass
class ResultSetResponse(betterproto.Message):
    """Response that contains a result set."""

    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    own_statement: bool = betterproto.bool_field(3)
    signature: "Signature" = betterproto.message_field(4)
    first_frame: "Frame" = betterproto.message_field(5)
    update_count: int = betterproto.uint64_field(6)
    # with no signature nor other data.
    metadata: "RpcMetadata" = betterproto.message_field(7)


@dataclass
class ExecuteResponse(betterproto.Message):
    """Response to PrepareAndExecuteRequest"""

    results: List["ResultSetResponse"] = betterproto.message_field(1)
    missing_statement: bool = betterproto.bool_field(2)
    metadata: "RpcMetadata" = betterproto.message_field(3)


@dataclass
class PrepareResponse(betterproto.Message):
    """Response to PrepareRequest"""

    statement: "StatementHandle" = betterproto.message_field(1)
    metadata: "RpcMetadata" = betterproto.message_field(2)


@dataclass
class FetchResponse(betterproto.Message):
    """Response to FetchRequest"""

    frame: "Frame" = betterproto.message_field(1)
    missing_statement: bool = betterproto.bool_field(2)
    missing_results: bool = betterproto.bool_field(3)
    metadata: "RpcMetadata" = betterproto.message_field(4)


@dataclass
class CreateStatementResponse(betterproto.Message):
    """Response to CreateStatementRequest"""

    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    metadata: "RpcMetadata" = betterproto.message_field(3)


@dataclass
class CloseStatementResponse(betterproto.Message):
    """Response to CloseStatementRequest"""

    metadata: "RpcMetadata" = betterproto.message_field(1)


@dataclass
class OpenConnectionResponse(betterproto.Message):
    """Response to OpenConnectionRequest {"""

    metadata: "RpcMetadata" = betterproto.message_field(1)


@dataclass
class CloseConnectionResponse(betterproto.Message):
    """Response to CloseConnectionRequest {"""

    metadata: "RpcMetadata" = betterproto.message_field(1)


@dataclass
class ConnectionSyncResponse(betterproto.Message):
    """Response to ConnectionSyncRequest"""

    conn_props: "ConnectionProperties" = betterproto.message_field(1)
    metadata: "RpcMetadata" = betterproto.message_field(2)


@dataclass
class DatabasePropertyElement(betterproto.Message):
    key: "DatabaseProperty" = betterproto.message_field(1)
    value: "TypedValue" = betterproto.message_field(2)
    metadata: "RpcMetadata" = betterproto.message_field(3)


@dataclass
class DatabasePropertyResponse(betterproto.Message):
    """Response for Meta#getDatabaseProperties()"""

    props: List["DatabasePropertyElement"] = betterproto.message_field(1)
    metadata: "RpcMetadata" = betterproto.message_field(2)


@dataclass
class ErrorResponse(betterproto.Message):
    """
    Send contextual information about some error over the wire from the server.
    """

    exceptions: List[str] = betterproto.string_field(1)
    has_exceptions: bool = betterproto.bool_field(7)
    error_message: str = betterproto.string_field(2)
    severity: "Severity" = betterproto.enum_field(3)
    error_code: int = betterproto.uint32_field(4)
    sql_state: str = betterproto.string_field(5)
    metadata: "RpcMetadata" = betterproto.message_field(6)


@dataclass
class SyncResultsResponse(betterproto.Message):
    missing_statement: bool = betterproto.bool_field(1)
    more_results: bool = betterproto.bool_field(2)
    metadata: "RpcMetadata" = betterproto.message_field(3)


@dataclass
class RpcMetadata(betterproto.Message):
    """Generic metadata for the server to return with each response."""

    server_address: str = betterproto.string_field(1)


@dataclass
class CommitResponse(betterproto.Message):
    """Response to a commit request"""

    pass


@dataclass
class RollbackResponse(betterproto.Message):
    """Response to a rollback request"""

    pass


@dataclass
class ExecuteBatchResponse(betterproto.Message):
    """Response to a batch update request"""

    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    update_counts: List[int] = betterproto.uint64_field(3)
    missing_statement: bool = betterproto.bool_field(4)
    metadata: "RpcMetadata" = betterproto.message_field(5)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: requests.proto
# plugin: python-betterproto
from .common_pb import *
from typing import Dict
import betterproto


@dataclass
class CatalogsRequest(betterproto.Message):
    """Request for Meta#getCatalogs()"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class DatabasePropertyRequest(betterproto.Message):
    """Request for Meta#getDatabaseProperties()"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class SchemasRequest(betterproto.Message):
    """
    Request for Meta#getSchemas(String, org.apache.calcite.avatica.Meta.Pat)}
    """

    catalog: str = betterproto.string_field(1)
    schema_pattern: str = betterproto.string_field(2)
    connection_id: str = betterproto.string_field(3)
    has_catalog: bool = betterproto.bool_field(4)
    has_schema_pattern: bool = betterproto.bool_field(5)


@dataclass
class TablesRequest(betterproto.Message):
    """
    Request for Request for Meta#getTables(String,
    org.apache.calcite.avatica.Meta.Pat,   org.apache.calcite.avatica.Meta.Pat,
    java.util.List)
    """

    catalog: str = betterproto.string_field(1)
    schema_pattern: str = betterproto.string_field(2)
    table_name_pattern: str = betterproto.string_field(3)
    type_list: List[str] = betterproto.string_field(4)
    has_type_list: bool = betterproto.bool_field(6)
    connection_id: str = betterproto.string_field(7)
    has_catalog: bool = betterproto.bool_field(8)
    has_schema_pattern: bool = betterproto.bool_field(9)
    has_table_name_pattern: bool = betterproto.bool_field(10)


@dataclass
class TableTypesRequest(betterproto.Message):
    """Request for Meta#getTableTypes()"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class ColumnsRequest(betterproto.Message):
    """
    Request for Meta#getColumns(String, org.apache.calcite.avatica.Meta.Pat,
    org.apache.calcite.avatica.Meta.Pat, org.apache.calcite.avatica.Meta.Pat).
    """

    catalog: str = betterproto.string_field(1)
    schema_pattern: str = betterproto.string_field(2)
    table_name_pattern: str = betterproto.string_field(3)
    column_name_pattern: str = betterproto.string_field(4)
    connection_id: str = betterproto.string_field(5)
    has_catalog: bool = betterproto.bool_field(6)
    has_schema_pattern: bool = betterproto.bool_field(7)
    has_table_name_pattern: bool = betterproto.bool_field(8)
    has_column_name_pattern: bool = betterproto.bool_field(9)


@dataclass
class TypeInfoRequest(betterproto.Message):
    """Request for Meta#getTypeInfo()"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class PrepareAndExecuteRequest(betterproto.Message):
    """
    Request for Meta#prepareAndExecute(Meta.StatementHandle, String, long,
    Meta.PrepareCallback)
    """

    connection_id: str = betterproto.string_field(1)
    sql: str = betterproto.string_field(2)
    max_row_count: int = betterproto.uint64_field(3)
    statement_id: int = betterproto.uint32_field(4)
    max_rows_total: int = betterproto.int64_field(5)
    first_frame_max_size: int = betterproto.int32_field(6)


@dataclass
class PrepareRequest(betterproto.Message):
    """Request for Meta.prepare(Meta.ConnectionHandle, String, long)"""

    connection_id: str = betterproto.string_field(1)
    sql: str = betterproto.string_field(2)
    max_row_count: int = betterproto.uint64_field(3)
    max_rows_total: int = betterproto.int64_field(4)


@dataclass
class FetchRequest(betterproto.Message):
    """Request for Meta#fetch(Meta.StatementHandle, List, long, int)"""

    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    offset: int = betterproto.uint64_field(3)
    fetch_max_row_count: int = betterproto.uint32_field(4)
    frame_max_size: int = betterproto.int32_field(5)


@dataclass
class CreateStatementRequest(betterproto.Message):
    """Request for Meta#createStatement(Meta.ConnectionHandle)"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class CloseStatementRequest(betterproto.Message):
    """Request for Meta#closeStatement(Meta.StatementHandle)"""

    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)


@dataclass
class OpenConnectionRequest(betterproto.Message):
    """
    Request for Meta#openConnection(Meta.ConnectionHandle, Map<String, String>)
    """

    connection_id: str = betterproto.string_field(1)
    info: Dict[str, str] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


@dataclass
class CloseConnectionRequest(betterproto.Message):
    """Request for Meta#closeConnection(Meta.ConnectionHandle)"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class ConnectionSyncRequest(betterproto.Message):
    connection_id: str = betterproto.string_field(1)
    conn_props: "ConnectionProperties" = betterproto.message_field(2)


@dataclass
class ExecuteRequest(betterproto.Message):
    """Request for Meta#execute(Meta.ConnectionHandle, list, long)"""

    statement_handle: "StatementHandle" = betterproto.message_field(1)
    parameter_values: List["TypedValue"] = betterproto.message_field(2)
    deprecated_first_frame_max_size: int = betterproto.uint64_field(3)
    has_parameter_values: bool = betterproto.bool_field(4)
    first_frame_max_size: int = betterproto.int32_field(5)


@dataclass
class SyncResultsRequest(betterproto.Message):
    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    state: "QueryState" = betterproto.message_field(3)
    offset: int = betterproto.uint64_field(4)


@dataclass
class CommitRequest(betterproto.Message):
    """Request to invoke a commit on a Connection"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class RollbackRequest(betterproto.Message):
    """Request to invoke rollback on a Connection"""

    connection_id: str = betterproto.string_field(1)


@dataclass
class PrepareAndExecuteBatchRequest(betterproto.Message):
    """Request to prepare and execute a collection of sql statements."""

    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    sql_commands: List[str] = betterproto.string_field(3)


@dataclass
class UpdateBatch(betterproto.Message):
    """Each command is a list of TypedValues"""

    parameter_values: List["TypedValue"] = betterproto.message_field(1)


@dataclass
class ExecuteBatchRequest(betterproto.Message):
    connection_id: str = betterproto.string_field(1)
    statement_id: int = betterproto.uint32_field(2)
    updates: List["UpdateBatch"] = betterproto.message_field(3)



import re
import asyncio
import aiohttp
import urllib.parse as urlparse
import betterproto
from aiophoenixdb import errors
from aiophoenixdb.avatica.proto import common_pb, requests_pb, responses_pb
from html.parser import HTMLParser
from aiohttp import BasicAuth, ClientError
from typing import Optional, Dict, TypeVar

_RESPONSE_MSG_JAVA_CLS_NAME = "org.apache.calcite.avatica.proto.Responses${cls_name}"
_REQUEST_MSG_JAVA_CLS_NAME = "org.apache.calcite.avatica.proto.Requests${cls_name}"


class JettyErrorPageParser(HTMLParser):

    def __init__(self):
        HTMLParser.__init__(self)
        self.path = []
        self.title = []
        self.message = []

    def handle_starttag(self, tag, attrs):
        self.path.append(tag)

    def handle_endtag(self, tag):
        self.path.pop()

    def handle_data(self, data):
        if len(self.path) > 2 and self.path[0] == 'html' and self.path[1] == 'body':
            if len(self.path) == 3 and self.path[2] == 'h2':
                self.title.append(data.strip())
            elif len(self.path) == 4 and self.path[2] == 'p' and self.path[3] == 'pre':
                self.message.append(data.strip())


def parse_url(url):
    url = urlparse.urlparse(url)
    if not url.scheme and not url.netloc and url.path:
        netloc = url.path
        if ':' not in netloc:
            netloc = '{}:8765'.format(netloc)

        return urlparse.ParseResult('http', netloc, '/', '', '', '')
    return url


# Defined in phoenix-core/src/main/java/org/apache/phoenix/exception/SQLExceptionCode.java
SQLSTATE_ERROR_CLASSES = [
    ('08', errors.OperationalError),  # Connection Exception
    ('22018', errors.IntegrityError),  # Constraint violatioin.
    ('22', errors.DataError),  # Data Exception
    ('23', errors.IntegrityError),  # Constraint Violation
    ('24', errors.InternalError),  # Invalid Cursor State
    ('25', errors.InternalError),  # Invalid Transaction State
    ('42', errors.ProgrammingError),  # Syntax Error or Access Rule Violation
    ('XLC', errors.OperationalError),  # Execution exceptions
    ('INT', errors.InternalError),  # Phoenix internal error
]


def raise_sql_error(code, sqlstate, message):
    for prefix, error_class in SQLSTATE_ERROR_CLASSES:
        if sqlstate.startswith(prefix):
            raise error_class(message, code, sqlstate)
    raise errors.InternalError(message, code, sqlstate)


def parse_and_raise_sql_error(message):
    match = re.findall(r'(?:([^ ]+): )?ERROR (\d+) \(([0-9A-Z]{5})\): (.*?) ->', message)
    if match is not None and len(match):
        exception, code, sqlstate, message = match[0]
        raise_sql_error(int(code), sqlstate, message)


def parse_error_page(html):
    parser = JettyErrorPageParser()
    parser.feed(html)
    if parser.title == ['HTTP ERROR: 500']:
        message = ' '.join(parser.message).strip()
        parse_and_raise_sql_error(message)
        raise errors.InternalError(message)


def parse_error_protobuf(text):
    try:
        message = common_pb.WireMessage()
        message.parse(text)

        err = responses_pb.ErrorResponse()
        if not err.parse(message.wrapped_message):
            raise Exception('No error message found')
    except Exception:
        # Not a protobuf error, fall through
        return

    parse_and_raise_sql_error(err.error_message)
    raise_sql_error(err.error_code, err.sql_state, err.error_message)
    # Not a protobuf error, fall through


_MESSAGE_TYPE = TypeVar("_MESSAGE_TYPE", bound=betterproto.Message)


class AvaticaClient(object):

    def __init__(self, url: str, max_retries: int, verify, extra_headers: Dict, auth: Optional[BasicAuth]):
        self._url = url
        self._headers = {'content-type': 'application/x-google-protobuf'}
        self._verify = verify
        self._max_retries = max_retries or 3
        if extra_headers:
            self._headers.update(extra_headers)

        self._session = aiohttp.ClientSession(
            headers=self._headers,
            auth=auth
        )

    async def __post_request(self, body, retry_delay=1):
        for _ in range(self._max_retries):
            request_args = {'data': body}
            if self._verify is not None:
                request_args.update(verify=self._verify)
            try:
                response = await self._session.post(self._url, **request_args)
            except ClientError:
                await asyncio.sleep(retry_delay)
            else:
                if response.status != 503:
                    return response
                else:
                    await asyncio.sleep(retry_delay)
        else:
            raise errors.MasRetriesError("Request retry more than the maximum number of attempts")

    async def _request(self, request_data,
                       expected_response_cls=None):
        request_name = request_data.__class__.__name__
        message = common_pb.WireMessage()
        message.name = _REQUEST_MSG_JAVA_CLS_NAME.format(cls_name=request_name)
        message.wrapped_message = request_data.SerializeToString()
        request_body = message.SerializeToString()

        response = await self.__post_request(request_body)
        response_body = await response.read()

        if response.status != 200:
            # logger.debug("Received response\n%s", response_body)
            if b'<html>' in response_body:
                parse_error_page(response_body.decode(response.get_encoding()))
            else:
                # assume the response is in protobuf format
                parse_error_protobuf(response_body)
            raise errors.InterfaceError('RPC request returned invalid status code', response.status)

        message = common_pb.WireMessage()
        message.parse(response_body)

        # logger.debug("Received response\n%s", message)

        if expected_response_cls is None:
            expected_response_type = request_name.replace('Request', 'Response')
        else:
            expected_response_type = expected_response_cls.__name__

        expected_response_type = _RESPONSE_MSG_JAVA_CLS_NAME.format(cls_name=expected_response_type)
        if message.name != expected_response_type:
            raise errors.InterfaceError(
                'unexpected response type "{}" expected "{}"'.format(message.name, expected_response_type))
        res: _MESSAGE_TYPE = expected_response_cls()
        res.parse(message.wrapped_message)
        return res

    async def get_catalogs(self, connection_id):
        request = requests_pb.CatalogsRequest()
        request.connection_id = connection_id
        response = await self._request(request, responses_pb.ResultSetResponse)
        return response

    async def get_schemas(self, connection_id,
                          catalog=None,
                          schema_pattern=None):
        request = requests_pb.SchemasRequest()
        request.connection_id = connection_id
        if catalog is not None:
            request.catalog = catalog
        if schema_pattern is not None:
            request.schema_pattern = schema_pattern
        response = await self._request(request, responses_pb.ResultSetResponse)
        return response

    async def get_tables(self, connection_id, catalog=None,
                         schema_pattern=None,
                         table_name_pattern=None,
                         type_list=None):
        request = requests_pb.TablesRequest()
        request.connection_id = connection_id
        if catalog is not None:
            request.catalog = catalog
        if schema_pattern is not None:
            request.schema_pattern = schema_pattern
        if table_name_pattern is not None:
            request.table_name_pattern = table_name_pattern
        if type_list is not None:
            request.type_list.extend(type_list)
        request.has_type_list = type_list is not None
        response = await self._request(request, responses_pb.ResultSetResponse)
        return response

    async def get_columns(self, connection_id,
                          catalog=None, schema_pattern=None,
                          table_name_pattern=None,
                          column_name_pattern=None):
        request = requests_pb.ColumnsRequest()
        request.connection_id = connection_id
        if catalog is not None:
            request.catalog = catalog
        if schema_pattern is not None:
            request.schema_pattern = schema_pattern
        if table_name_pattern is not None:
            request.table_name_pattern = table_name_pattern
        if column_name_pattern is not None:
            request.column_name_pattern = column_name_pattern
        response = await self._request(request, responses_pb.ResultSetResponse)
        return response

    async def get_table_types(self, connection_id):
        request = requests_pb.TableTypesRequest()
        request.connection_id = connection_id
        response = await self._request(request, responses_pb.ResultSetResponse)
        return response

    async def get_type_info(self, connection_id):
        request = requests_pb.TypeInfoRequest()
        request.connection_id = connection_id
        response = await self._request(request, responses_pb.ResultSetResponse)
        return response

    async def get_sync_results(self, connection_id, statement_id, state):
        request = requests_pb.SyncResultsRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        # todo copyFrom
        request.state.parse(state)
        response = await self._request(request, responses_pb.SyncResultsResponse)
        return response

    async def connection_sync_dict(self, connection_id, conn_props=None):
        conn_props = await self.connection_sync(connection_id, conn_props)
        return {
            'autoCommit': conn_props.auto_commit,
            'readOnly': conn_props.read_only,
            'transactionIsolation': conn_props.transaction_isolation,
            'catalog': conn_props.catalog,
            'schema': conn_props.schema}

    async def connection_sync(self, connection_id, conn_props=None):
        """Synchronizes connection properties with the server.

        :param connection_id:
            ID of the current connection.

        :param conn_props:
            Dictionary with the properties that should be changed.

        :returns:
            A ``common_pb2.ConnectionProperties`` object.
        """
        if conn_props:
            props = conn_props.copy()
        else:
            props = {}

        request = requests_pb.ConnectionSyncRequest()
        request.connection_id = connection_id
        request.conn_props.has_auto_commit = True
        request.conn_props.has_read_only = True
        if 'autoCommit' in props:
            request.conn_props.auto_commit = props.pop('autoCommit')
        if 'readOnly' in props:
            request.conn_props.read_only = props.pop('readOnly')
        if 'transactionIsolation' in props:
            request.conn_props.transaction_isolation = props.pop('transactionIsolation', None)
        if 'catalog' in props:
            request.conn_props.catalog = props.pop('catalog', None)
        if 'schema' in props:
            request.conn_props.schema = props.pop('schema', None)

        # if props:
        #     logger.warning("Unhandled connection property:" + props)

        # response_data = await self._apply(request)
        # response = ConnectionSyncResponse()
        # response.parse(response_data)

        response = await self._request(request, responses_pb.ConnectionSyncResponse)
        return response.conn_props

    async def open_connection(self, connection_id, info=None):
        """Opens a new connection.

        :param info:
        :param connection_id:
            ID of the connection to open.
        """
        request = requests_pb.OpenConnectionRequest()
        request.connection_id = connection_id
        if info is not None:
            # Info is a list of repeated pairs, setting a dict directly fails
            for k, v in info.items():
                request.info[k] = v

        # response_data = await self._apply(request)
        # response = responses_pb.OpenConnectionResponse()
        # response.parse(response_data)
        return await self._request(request, responses_pb.OpenConnectionResponse)

    async def close_connection(self, connection_id):
        """Closes a connection.

        :param connection_id:
            ID of the connection to close.
        """
        request = requests_pb.CloseConnectionRequest()
        request.connection_id = connection_id
        return await self._request(request, responses_pb.CloseConnectionResponse)

    async def create_statement(self, connection_id):
        """Creates a new statement.

        :param connection_id:
            ID of the current connection.

        :returns:
            New statement ID.
        """
        request = requests_pb.CreateStatementRequest()
        request.connection_id = connection_id

        response = await self._request(request, responses_pb.CreateStatementResponse)
        return response.statement_id

    async def close_statement(self, connection_id, statement_id):
        """Closes a statement.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to close.
        """
        request = requests_pb.CloseStatementRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id

        return await self._request(request, responses_pb.CloseStatementResponse)

    async def prepare_and_execute(self, connection_id, statement_id, sql, max_rows_total=None,
                                  first_frame_max_size=None):
        """Prepares and immediately executes a statement.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to prepare.

        :param sql:
            SQL query.

        :param max_rows_total:
            The maximum number of rows that will be allowed for this query.

        :param first_frame_max_size:
            The maximum number of rows that will be returned in the first Frame returned for this query.

        :returns:
            Result set with the signature of the prepared statement and the first frame data.
        """
        request = requests_pb.PrepareAndExecuteRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        request.sql = sql
        if max_rows_total is not None:
            request.max_rows_total = max_rows_total
        if first_frame_max_size is not None:
            request.first_frame_max_size = first_frame_max_size
        # 提交request, 发起请求
        response = await self._request(request, responses_pb.ExecuteResponse)
        return response.results

    async def prepare(self, connection_id, sql, max_rows_total=None):
        """Prepares a statement.

        :param connection_id:
            ID of the current connection.

        :param sql:
            SQL query.

        :param max_rows_total:
            The maximum number of rows that will be allowed for this query.

        :returns:
            Signature of the prepared statement.
        """
        request = requests_pb.PrepareRequest()
        request.connection_id = connection_id
        request.sql = sql
        if max_rows_total is not None:
            request.max_rows_total = max_rows_total

        response = await self._request(request, responses_pb.PrepareResponse)
        return response.statement

    async def execute(self, connection_id,
                      statement_id,
                      signature,
                      parameter_values=None,
                      first_frame_max_size=None):
        """Returns a frame of rows.

        The frame describes whether there may be another frame. If there is not
        another frame, the current iteration is done when we have finished the
        rows in this frame.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to fetch rows from.

        :param signature:
            common_pb2.Signature object

        :param parameter_values:
            A list of parameter values, if statement is to be executed; otherwise ``None``.

        :param first_frame_max_size:
            The maximum number of rows that will be returned in the first Frame returned for this query.

        :returns:
            Frame data, or ``None`` if there are no more.
        """
        request = requests_pb.ExecuteRequest()
        request.statement_handle.id = statement_id
        request.statement_handle.connection_id = connection_id
        # todo copy from
        request.statement_handle.signature = signature
        if parameter_values is not None:
            request.parameter_values.extend(parameter_values)
            request.has_parameter_values = True
        if first_frame_max_size is not None:
            request.deprecated_first_frame_max_size = first_frame_max_size
            request.first_frame_max_size = first_frame_max_size

        response = await self._request(request, responses_pb.ExecuteResponse)
        return response.results

    async def execute_batch(self, connection_id, statement_id, rows):
        """Returns an array of update counts corresponding to each row written.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to fetch rows from.

        :param rows:
            A list of lists corresponding to the columns to bind to the statement
            for many rows.

        :returns:
            Update counts for the writes.
        """
        request = requests_pb.ExecuteBatchRequest()
        request.statement_id = statement_id
        request.connection_id = connection_id
        if rows is not None:
            for row in rows:
                batch = requests_pb.UpdateBatch()
                for col in row:
                    batch.parameter_values.append(col)
                request.updates.append(batch)

        response = await self._request(request, responses_pb.ExecuteBatchResponse)
        if response.missing_statement:
            raise errors.DatabaseError('ExecuteBatch reported missing statement', -1)
        return response.update_counts

    async def fetch(self, connection_id, statement_id, offset=0, frame_max_size=None):
        """Returns a frame of rows.

        The frame describes whether there may be another frame. If there is not
        another frame, the current iteration is done when we have finished the
        rows in this frame.

        :param connection_id:
            ID of the current connection.

        :param statement_id:
            ID of the statement to fetch rows from.

        :param offset:
            Zero-based offset of first row in the requested frame.

        :param frame_max_size:
            Maximum number of rows to return; negative means no limit.

        :returns:
            Frame data, or ``None`` if there are no more.
        """
        request = requests_pb.FetchRequest()
        request.connection_id = connection_id
        request.statement_id = statement_id
        request.offset = offset
        if frame_max_size is not None:
            request.frame_max_size = frame_max_size

        response = await self._request(request, responses_pb.FetchResponse)
        return response.frame

    async def commit(self, connection_id):
        """TODO Commits the transaction

        :param connection_id:
            ID of the current connection.
        """
        request = requests_pb.CommitRequest()
        request.connection_id = connection_id
        return await self._request(request)

    async def rollback(self, connection_id):
        """TODO Rolls back the transaction

        :param connection_id:
            ID of the current connection.
        """
        request = requests_pb.RollbackRequest()
        request.connection_id = connection_id
        return await self._request(request)

    async def close(self):
        await self._session.close()

    def __enter__(self):
        raise TypeError("Use async with instead")

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


if __name__ == '__main__':
    pass



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, List

from aiophoenixdb.connection import Connection
from aiophoenixdb.avatica.proto.common_pb import MetaDataOperationArgument, ColumnMetaData

__all__ = ['Meta']

logger: logging.Logger


class Meta(object):
    """Database meta for querying MetaData
    """

    _connection: Connection

    def __init__(self, connection: Connection): ...

    def get_catalogs(self) -> Any: ...

    def get_schemas(self, catalog=None, schema_pattern=None) -> Any: ...

    def get_tables(self, catalog=None, schema_pattern=None, table_name_pattern=None, type_list=None) -> Any: ...

    def get_columns(self, catalog=None, schema_pattern=None, table_name_pattern=None, column_name_pattern=None) -> Any: ...

    def get_table_types(self) -> Any: ...

    def get_type_info(self) -> Any: ...

    async def get_primary_keys(self, catalog=None, schema=None, table=None) -> List[Any]: ...

    async def get_index_info(self, catalog=None, schema=None, table=None, unique=False, approximate=False) -> List[Any]: ...

    @staticmethod
    def _column_meta_data_factory(ordinal: int, column_name: str, jdbc_code: int) -> ColumnMetaData: ...

    @staticmethod
    def _moa_string_arg_factory(arg: str) -> MetaDataOperationArgument: ...

    @staticmethod
    def _moa_bool_arg_factory(arg: bool) -> MetaDataOperationArgument: ...

    @staticmethod
    def _fix_default(rows: List[Any], catalog=None, schema_pattern=None) -> Any: ...



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import betterproto
from typing import List
from .common_pb import Signature
from .common_pb import Frame
from .common_pb import StatementHandle
from .common_pb import Severity
from .common_pb import DatabaseProperty
from .common_pb import TypedValue
from .common_pb import ConnectionProperties


class ResultSetResponse(betterproto.Message):
    """Response that contains a result set."""

    connection_id: str
    statement_id: int
    own_statement: bool
    signature: "Signature"
    first_frame: "Frame"
    update_count: int
    # with no signature nor other data.
    metadata: "RpcMetadata"


class ExecuteResponse(betterproto.Message):
    """Response to PrepareAndExecuteRequest"""

    results: List["ResultSetResponse"]
    missing_statement: bool
    metadata: "RpcMetadata"


class PrepareResponse(betterproto.Message):
    """Response to PrepareRequest"""

    statement: "StatementHandle"
    metadata: "RpcMetadata"


class FetchResponse(betterproto.Message):
    """Response to FetchRequest"""

    frame: "Frame"
    missing_statement: bool
    missing_results: bool
    metadata: "RpcMetadata"


class CreateStatementResponse(betterproto.Message):
    """Response to CreateStatementRequest"""

    connection_id: str
    statement_id: int
    metadata: "RpcMetadata"


class CloseStatementResponse(betterproto.Message):
    """Response to CloseStatementRequest"""

    metadata: "RpcMetadata"


class OpenConnectionResponse(betterproto.Message):
    """Response to OpenConnectionRequest {"""

    metadata: "RpcMetadata"


class CloseConnectionResponse(betterproto.Message):
    """Response to CloseConnectionRequest {"""

    metadata: "RpcMetadata"


class ConnectionSyncResponse(betterproto.Message):
    """Response to ConnectionSyncRequest"""

    conn_props: "ConnectionProperties"
    metadata: "RpcMetadata"


class DatabasePropertyElement(betterproto.Message):
    key: "DatabaseProperty"
    value: "TypedValue"
    metadata: "RpcMetadata"


class DatabasePropertyResponse(betterproto.Message):
    """Response for Meta#getDatabaseProperties()"""

    props: List["DatabasePropertyElement"]
    metadata: "RpcMetadata"


class ErrorResponse(betterproto.Message):
    """
    Send contextual information about some error over the wire from the server.
    """

    exceptions: List[str]
    has_exceptions: bool
    error_message: str
    severity: "Severity"
    error_code: int
    sql_state: str
    metadata: "RpcMetadata"


class SyncResultsResponse(betterproto.Message):
    missing_statement: bool
    more_results: bool
    metadata: "RpcMetadata"


class RpcMetadata(betterproto.Message):
    """Generic metadata for the server to return with each response."""

    server_address: str


class CommitResponse(betterproto.Message):
    """Response to a commit request"""

    pass


class RollbackResponse(betterproto.Message):
    """Response to a rollback request"""
    pass


class ExecuteBatchResponse(betterproto.Message):
    """Response to a batch update request"""

    connection_id: str
    statement_id: int
    update_counts: List[int]
    missing_statement: bool
    metadata: "RpcMetadata"



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import betterproto
from typing import Dict, List
from .common_pb import ConnectionProperties
from .common_pb import StatementHandle
from .common_pb import TypedValue
from .common_pb import QueryState


class CatalogsRequest(betterproto.Message):
    """Request for Meta#getCatalogs()"""

    connection_id: str


class DatabasePropertyRequest(betterproto.Message):
    """Request for Meta#getDatabaseProperties()"""

    connection_id: str


class SchemasRequest(betterproto.Message):
    """
    Request for Meta#getSchemas(String, org.apache.calcite.avatica.Meta.Pat)}
    """

    catalog: str
    schema_pattern: str
    connection_id: str
    has_catalog: bool
    has_schema_pattern: bool


class TablesRequest(betterproto.Message):
    """
    Request for Request for Meta#getTables(String,
    org.apache.calcite.avatica.Meta.Pat,   org.apache.calcite.avatica.Meta.Pat,
    java.util.List)
    """

    catalog: str
    schema_pattern: str
    table_name_pattern: str
    type_list: List[str]
    has_type_list: bool
    connection_id: str
    has_catalog: bool
    has_schema_pattern: bool
    has_table_name_pattern: bool


class TableTypesRequest(betterproto.Message):
    """Request for Meta#getTableTypes()"""

    connection_id: str


class ColumnsRequest(betterproto.Message):
    """
    Request for Meta#getColumns(String, org.apache.calcite.avatica.Meta.Pat,
    org.apache.calcite.avatica.Meta.Pat, org.apache.calcite.avatica.Meta.Pat).
    """

    catalog: str
    schema_pattern: str
    table_name_pattern: str
    column_name_pattern: str
    connection_id: str
    has_catalog: bool
    has_schema_pattern: bool
    has_table_name_pattern: bool
    has_column_name_pattern: bool


class TypeInfoRequest(betterproto.Message):
    """Request for Meta#getTypeInfo()"""

    connection_id: str


class PrepareAndExecuteRequest(betterproto.Message):
    """
    Request for Meta#prepareAndExecute(Meta.StatementHandle, String, long,
    Meta.PrepareCallback)
    """

    connection_id: str
    sql: str
    max_row_count: int
    statement_id: int
    max_rows_total: int
    first_frame_max_size: int


class PrepareRequest(betterproto.Message):
    """Request for Meta.prepare(Meta.ConnectionHandle, String, long)"""

    connection_id: str
    sql: str
    max_row_count: int
    max_rows_total: int


class FetchRequest(betterproto.Message):
    """Request for Meta#fetch(Meta.StatementHandle, List, long, int)"""

    connection_id: str
    statement_id: int
    offset: int
    fetch_max_row_count: int
    frame_max_size: int


class CreateStatementRequest(betterproto.Message):
    """Request for Meta#createStatement(Meta.ConnectionHandle)"""

    connection_id: str


class CloseStatementRequest(betterproto.Message):
    """Request for Meta#closeStatement(Meta.StatementHandle)"""

    connection_id: str
    statement_id: int


class OpenConnectionRequest(betterproto.Message):
    """
    Request for Meta#openConnection(Meta.ConnectionHandle, Map<String, String>)
    """

    connection_id: str
    info: Dict[str, str] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


class CloseConnectionRequest(betterproto.Message):
    """Request for Meta#closeConnection(Meta.ConnectionHandle)"""

    connection_id: str


class ConnectionSyncRequest(betterproto.Message):
    connection_id: str
    conn_props: "ConnectionProperties"


class ExecuteRequest(betterproto.Message):
    """Request for Meta#execute(Meta.ConnectionHandle, list, long)"""

    statement_handle: "StatementHandle"
    parameter_values: List["TypedValue"]
    deprecated_first_frame_max_size: int
    has_parameter_values: bool
    first_frame_max_size: int


class SyncResultsRequest(betterproto.Message):
    connection_id: str
    statement_id: int
    state: "QueryState"
    offset: int


class CommitRequest(betterproto.Message):
    """Request to invoke a commit on a Connection"""

    connection_id: str


class RollbackRequest(betterproto.Message):
    """Request to invoke rollback on a Connection"""

    connection_id: str


class PrepareAndExecuteBatchRequest(betterproto.Message):
    """Request to prepare and execute a collection of sql statements."""

    connection_id: str
    statement_id: int
    sql_commands: List[str]


class UpdateBatch(betterproto.Message):
    """Each command is a list of TypedValues"""

    parameter_values: List["TypedValue"]


class ExecuteBatchRequest(betterproto.Message):
    connection_id: str
    statement_id: int
    updates: List["UpdateBatch"]



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging

from aiophoenixdb.avatica.proto import common_pb
from aiophoenixdb.errors import ProgrammingError
from aiophoenixdb.cursors import DictCursor


__all__ = ['Meta']

logger = logging.getLogger(__name__)


class Meta(object):
    """Database meta for querying MetaData
    """

    def __init__(self, connection):
        self._connection = connection

    async def get_catalogs(self):
        if self._connection.closed:
            raise ProgrammingError('The connection is already closed.')
        result = await self._connection.client.get_catalogs(self._connection.connect_id)
        async with DictCursor(self._connection) as cursor:
            await cursor.process_result(result)
            return cursor.fetchall()

    async def get_schemas(self, catalog=None, schema_pattern=None):
        if self._connection.closed:
            raise ProgrammingError('The connection is already closed.')
        result = await self._connection.client.get_schemas(self._connection.connect_id, catalog, schema_pattern)
        async with DictCursor(self._connection) as cursor:
            await cursor.process_result(result)
            return self._fix_default(await cursor.fetchall(), schema_pattern=schema_pattern)

    async def get_tables(self, catalog=None, schema_pattern=None, table_name_pattern=None, type_list=None):
        if self._connection.closed:
            raise ProgrammingError('The connection is already closed.')
        result = await self._connection.client.get_tables(
            self._connection.connect_id, catalog, schema_pattern, table_name_pattern, type_list=type_list)
        async with DictCursor(self._connection) as cursor:
            await cursor.process_result(result)
            return self._fix_default(await cursor.fetchall(), catalog, schema_pattern)

    async def get_columns(self, catalog=None, schema_pattern=None, table_name_pattern=None,
                          column_name_pattern=None):
        if self._connection.closed:
            raise ProgrammingError('The connection is already closed.')
        result = await self._connection.client.get_columns(
            self._connection.connect_id, catalog, schema_pattern, table_name_pattern, column_name_pattern)
        async with DictCursor(self._connection) as cursor:
            await cursor.process_result(result)
            return self._fix_default(await cursor.fetchall(), catalog, schema_pattern)

    async def get_table_types(self):
        if self._connection.closed:
            raise ProgrammingError('The connection is already closed.')
        result = await self._connection.client.get_table_types(self._connection.connect_id)
        async with DictCursor(self._connection) as cursor:
            await cursor.process_result(result)
            return await cursor.fetchall()

    async def get_type_info(self):
        if self._connection.closed:
            raise ProgrammingError('The connection is already closed.')
        result = await self._connection.client.get_type_info(self._connection.connect_id)
        async with DictCursor(self._connection) as cursor:
            await cursor.process_result(result)
            return await cursor.fetchall()

    async def get_primary_keys(self, catalog=None, schema=None, table=None):
        if self._connection.closed:
            raise ProgrammingError('The cursor is already closed.')

        state = common_pb.QueryState()
        state.type = common_pb.StateType.METADATA
        state.op = common_pb.MetaDataOperation.GET_PRIMARY_KEYS
        state.has_args = True
        state.has_op = True

        catalog_arg = self._moa_string_arg_factory(catalog)
        schema_arg = self._moa_string_arg_factory(schema)
        table_arg = self._moa_string_arg_factory(table)
        state.args.extend([catalog_arg, schema_arg, table_arg])

        async with DictCursor(self._connection) as cursor:
            sync_result_response = await cursor.get_sync_results(state)
            if not sync_result_response.more_results:
                return []

            signature = common_pb.Signature()
            signature.columns.append(self._column_meta_data_factory(1, 'TABLE_CAT', 12))
            signature.columns.append(self._column_meta_data_factory(2, 'TABLE_SCHEM', 12))
            signature.columns.append(self._column_meta_data_factory(3, 'TABLE_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(4, 'COLUMN_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(5, 'KEY_SEQ', 5))
            signature.columns.append(self._column_meta_data_factory(6, 'PK_NAME', 12))
            # The following are non-standard Phoenix extensions
            # This returns '\x00\x00\x00A' or '\x00\x00\x00D' , but that's consistent with Java
            signature.columns.append(self._column_meta_data_factory(7, 'ASC_OR_DESC', 12))
            signature.columns.append(self._column_meta_data_factory(8, 'DATA_TYPE', 5))
            signature.columns.append(self._column_meta_data_factory(9, 'TYPE_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(10, 'COLUMN_SIZE', 5))
            signature.columns.append(self._column_meta_data_factory(11, 'TYPE_ID', 5))
            signature.columns.append(self._column_meta_data_factory(12, 'VIEW_CONSTANT', 12))

            await cursor.fetch(signature)
            return await cursor.fetchall()

    async def get_index_info(self, catalog=None, schema=None, table=None, unique=False, approximate=False):
        if self._connection.closed:
            raise ProgrammingError('The cursor is already closed.')

        state = common_pb.QueryState()
        state.type = common_pb.StateType.METADATA
        state.op = common_pb.MetaDataOperation.GET_INDEX_INFO
        state.has_args = True
        state.has_op = True

        catalog_arg = self._moa_string_arg_factory(catalog)
        schema_arg = self._moa_string_arg_factory(schema)
        table_arg = self._moa_string_arg_factory(table)
        unique_arg = self._moa_bool_arg_factory(unique)
        approximate_arg = self._moa_bool_arg_factory(approximate)

        state.args.extend([catalog_arg, schema_arg, table_arg, unique_arg, approximate_arg])

        async with DictCursor(self._connection) as cursor:
            sync_result_response = await cursor.get_sync_results(state)
            if not sync_result_response.more_results:
                return []

            signature = common_pb.Signature()
            signature.columns.append(self._column_meta_data_factory(1, 'TABLE_CAT', 12))
            signature.columns.append(self._column_meta_data_factory(2, 'TABLE_SCHEM', 12))
            signature.columns.append(self._column_meta_data_factory(3, 'TABLE_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(4, 'NON_UNIQUE', 16))
            signature.columns.append(self._column_meta_data_factory(5, 'INDEX_QUALIFIER', 12))
            signature.columns.append(self._column_meta_data_factory(6, 'INDEX_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(7, 'TYPE', 5))
            signature.columns.append(self._column_meta_data_factory(8, 'ORDINAL_POSITION', 5))
            signature.columns.append(self._column_meta_data_factory(9, 'COLUMN_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(10, 'ASC_OR_DESC', 12))
            signature.columns.append(self._column_meta_data_factory(11, 'CARDINALITY', 5))
            signature.columns.append(self._column_meta_data_factory(12, 'PAGES', 5))
            signature.columns.append(self._column_meta_data_factory(13, 'FILTER_CONDITION', 12))
            # The following are non-standard Phoenix extensions
            signature.columns.append(self._column_meta_data_factory(14, 'DATA_TYPE', 5))
            signature.columns.append(self._column_meta_data_factory(15, 'TYPE_NAME', 12))
            signature.columns.append(self._column_meta_data_factory(16, 'TYPE_ID', 5))
            signature.columns.append(self._column_meta_data_factory(17, 'COLUMN_FAMILY', 12))
            signature.columns.append(self._column_meta_data_factory(18, 'COLUMN_SIZE', 5))
            signature.columns.append(self._column_meta_data_factory(19, 'ARRAY_SIZE', 5))

            await cursor.fetch(signature)
            return await cursor.fetchall()

    @staticmethod
    def _column_meta_data_factory(ordinal, column_name, jdbc_code):
        cmd = common_pb.ColumnMetaData()
        cmd.ordinal = ordinal
        cmd.column_name = column_name
        cmd.type.id = jdbc_code
        cmd.nullable = 2
        return cmd

    @staticmethod
    def _moa_string_arg_factory(arg):
        moa = common_pb.MetaDataOperationArgument()
        if arg is None:
            moa.type = common_pb.MetaDataOperationArgument.ArgumentType.NULL
        else:
            moa.type = common_pb.MetaDataOperationArgument.ArgumentType.STRING
            moa.string_value = arg
        return moa

    @staticmethod
    def _moa_bool_arg_factory(arg):
        moa = common_pb.MetaDataOperationArgument()
        if arg is None:
            moa.type = common_pb.MetaDataOperationArgument.ArgumentType.NULL
        else:
            moa.type = common_pb.MetaDataOperationArgument.ArgumentType.BOOL
            moa.bool_value = arg
        return moa

    @staticmethod
    def _fix_default(rows, catalog=None, schema_pattern=None):
        """Workaround for PHOENIX-6003"""
        if schema_pattern == '':
            rows = [row for row in rows if row['TABLE_SCHEM'] is None]
        if catalog == '':
            rows = [row for row in rows if row['TABLE_CATALOG'] is None]
        # Couldn't find a sane way to do it that works on 2 and 3
        if sys.version_info.major == 3:
            return [{k: v or '' for k, v in row.items()} for row in rows]
        else:
            return [{k: v or '' for k, v in row.iteritems()} for row in rows]



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
from aiophoenixdb import errors
from aiophoenixdb.errors import ProgrammingError
from aiophoenixdb.meta import Meta


__all__ = ['Connection']

logger = logging.getLogger(__name__)

AVATICA_PROPERTIES = ('autoCommit', 'autocommit', 'readOnly', 'readonly', 'transactionIsolation',
                      'catalog', 'schema')


class Connection(object):
    """Database connection.

    You should not construct this object manually, use :func:`~aiophoenixdb.connect` instead.
    """

    """
    The default cursor factory used by :meth:`cursor` if the parameter is not specified.
    """

    def __init__(self, client, cursor_factory=None, **kwargs):
        self._client = client
        self._closed = False
        if cursor_factory is not None:
            self.cursor_factory = cursor_factory
        else:
            from aiophoenixdb.cursors import Cursor
            self.cursor_factory = Cursor
        self._cursors = []
        self._phoenix_props, self.avatica_props_init = Connection._map_conn_props(kwargs)
        self._conn_id = str(uuid.uuid4())
        self._avatica_props = dict()

    async def connect(self):
        await self.open()
        await self.set_session(**self.avatica_props_init)

    # def __del__(self):
    #     asyncio.get_running_loop().run_until_complete(self.close())
    #     print("connect del del del del ", self._closed)

    def __enter__(self):
        raise TypeError("Use async with instead")

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self._closed:
            await self.close()

    @property
    def client(self):
        return self._client

    @property
    def closed(self):
        return self._closed

    @property
    def connect_id(self):
        return self._conn_id

    @property
    def _default_avatica_props(self):
        return {'autoCommit': False,
                'readOnly': False,
                'transactionIsolation': 0,
                'catalog': '',
                'schema': ''}

    @staticmethod
    def _map_conn_props(conn_props):
        """Sorts and prepocesses args that should be passed to Phoenix and Avatica"""

        avatica_props = dict([(k, conn_props[k]) for k in conn_props.keys() if k in AVATICA_PROPERTIES])
        phoenix_props = dict([(k, conn_props[k]) for k in conn_props.keys() if k not in AVATICA_PROPERTIES])
        avatica_props = Connection._map_legacy_avatica_props(avatica_props)

        return phoenix_props, avatica_props

    @staticmethod
    def _map_legacy_avatica_props(props):
        if 'autocommit' in props:
            props['autoCommit'] = bool(props.pop('autocommit'))
        if 'readonly' in props:
            props['readOnly'] = bool(props.pop('readonly'))
        return props

    async def open(self):
        """Opens the connection."""
        await self._client.open_connection(self._conn_id, info=self._phoenix_props)

    async def close(self):
        """Closes the connection.
        No further operations are allowed, either on the connection or any
        of its cursors, once the connection is closed.

        If the connection is used in a ``with`` statement, this method will
        be automatically called at the end of the ``with`` block.
        """
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        for cursor_ref in self._cursors:
            cursor = cursor_ref()
            if cursor is not None and not cursor.closed:
                await cursor.close()
        await self._client.close_connection(self._conn_id)
        await self._client.close()
        self._closed = True

    async def commit(self):
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        await self._client.commit(self._conn_id)

    async def rollback(self):
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        await self._client.rollback(self._conn_id)

    def cursor(self, cursor_factory=None):
        """Creates a new cursor.

        :param cursor_factory:
            This argument can be used to create non-standard cursors.
            The class returned must be a subclass of
            :class:`~aiophoenixdb.cursor.Cursor` (for example :class:`~aiophoenixdb.cursor.DictCursor`).
            A default factory for the connection can also be specified using the
            :attr:`cursor_factory` attribute.

        :returns:
            A :class:`~aiophoenixdb.cursor.Cursor` object.
        """
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        cursor = (cursor_factory or self.cursor_factory)(self)
        self._cursors.append(cursor.ref(self._cursors.remove))
        return cursor

    async def set_session(self, **props):
        """Sets one or more parameters in the current connection.

        :param autocommit:
            Switch the connection to autocommit mode.

        :param readonly:
            Switch the connection to read-only mode.
        """
        props = Connection._map_legacy_avatica_props(props)
        self._avatica_props = await self._client.connection_sync_dict(self._conn_id, props)

    @property
    def autocommit(self):
        """Read/write attribute for switching the connection's autocommit mode."""
        return self._avatica_props['autoCommit']

    async def set_autocommit(self, value):
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        self._avatica_props = await self._client.connection_sync_dict(self._conn_id, {'autoCommit': bool(value)})

    @property
    def readonly(self):
        """Read/write attribute for switching the connection's readonly mode."""
        return self._avatica_props['readOnly']

    async def set_readonly(self, value):
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        self._avatica_props = await self._client.connection_sync_dict(self._conn_id, {'readOnly': bool(value)})

    @property
    def transactionisolation(self):
        return self._avatica_props['_transactionIsolation']

    async def set_transactionisolation(self, value):
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        self._avatica_props = await self._client.connection_sync_dict(
            self._conn_id, {'transactionIsolation': bool(value)})

    def meta(self):
        """Creates a new meta.

        :returns:
            A :class:`~aiophoenixdb.meta` object.
        """
        if self._closed:
            raise ProgrammingError('The connection is already closed.')
        meta = Meta(self)
        return meta


for name in errors.__all__:
    setattr(Connection, name, getattr(errors, name))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: common.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import List
import betterproto


class StatementType(betterproto.Enum):
    """Has to be consistent with Meta.StatementType"""

    SELECT = 0
    INSERT = 1
    UPDATE = 2
    DELETE = 3
    UPSERT = 4
    MERGE = 5
    OTHER_DML = 6
    CREATE = 7
    DROP = 8
    ALTER = 9
    OTHER_DDL = 10
    CALL = 11


class Rep(betterproto.Enum):
    PRIMITIVE_BOOLEAN = 0
    PRIMITIVE_BYTE = 1
    PRIMITIVE_CHAR = 2
    PRIMITIVE_SHORT = 3
    PRIMITIVE_INT = 4
    PRIMITIVE_LONG = 5
    PRIMITIVE_FLOAT = 6
    PRIMITIVE_DOUBLE = 7
    BOOLEAN = 8
    BYTE = 9
    CHARACTER = 10
    SHORT = 11
    INTEGER = 12
    LONG = 13
    FLOAT = 14
    DOUBLE = 15
    BIG_INTEGER = 25
    BIG_DECIMAL = 26
    JAVA_SQL_TIME = 16
    JAVA_SQL_TIMESTAMP = 17
    JAVA_SQL_DATE = 18
    JAVA_UTIL_DATE = 19
    BYTE_STRING = 20
    STRING = 21
    NUMBER = 22
    OBJECT = 23
    NULL = 24
    ARRAY = 27
    STRUCT = 28
    MULTISET = 29


class Severity(betterproto.Enum):
    """
    The severity of some unexpected outcome to an operation. Protobuf enum
    values must be unique across all other enums
    """

    UNKNOWN_SEVERITY = 0
    FATAL_SEVERITY = 1
    ERROR_SEVERITY = 2
    WARNING_SEVERITY = 3


class MetaDataOperation(betterproto.Enum):
    """Enumeration corresponding to DatabaseMetaData operations"""

    GET_ATTRIBUTES = 0
    GET_BEST_ROW_IDENTIFIER = 1
    GET_CATALOGS = 2
    GET_CLIENT_INFO_PROPERTIES = 3
    GET_COLUMN_PRIVILEGES = 4
    GET_COLUMNS = 5
    GET_CROSS_REFERENCE = 6
    GET_EXPORTED_KEYS = 7
    GET_FUNCTION_COLUMNS = 8
    GET_FUNCTIONS = 9
    GET_IMPORTED_KEYS = 10
    GET_INDEX_INFO = 11
    GET_PRIMARY_KEYS = 12
    GET_PROCEDURE_COLUMNS = 13
    GET_PROCEDURES = 14
    GET_PSEUDO_COLUMNS = 15
    GET_SCHEMAS = 16
    GET_SCHEMAS_WITH_ARGS = 17
    GET_SUPER_TABLES = 18
    GET_SUPER_TYPES = 19
    GET_TABLE_PRIVILEGES = 20
    GET_TABLES = 21
    GET_TABLE_TYPES = 22
    GET_TYPE_INFO = 23
    GET_UDTS = 24
    GET_VERSION_COLUMNS = 25


class StateType(betterproto.Enum):
    SQL = 0
    METADATA = 1


class CursorFactoryStyle(betterproto.Enum):
    OBJECT = 0
    RECORD = 1
    RECORD_PROJECTION = 2
    ARRAY = 3
    LIST = 4
    MAP = 5


class MetaDataOperationArgumentArgumentType(betterproto.Enum):
    STRING = 0
    BOOL = 1
    INT = 2
    REPEATED_STRING = 3
    REPEATED_INT = 4
    NULL = 5


@dataclass
class ConnectionProperties(betterproto.Message):
    """Details about a connection"""

    is_dirty: bool = betterproto.bool_field(1)
    auto_commit: bool = betterproto.bool_field(2)
    has_auto_commit: bool = betterproto.bool_field(7)
    read_only: bool = betterproto.bool_field(3)
    has_read_only: bool = betterproto.bool_field(8)
    transaction_isolation: int = betterproto.uint32_field(4)
    catalog: str = betterproto.string_field(5)
    schema: str = betterproto.string_field(6)


@dataclass
class StatementHandle(betterproto.Message):
    """Statement handle"""

    connection_id: str = betterproto.string_field(1)
    id: int = betterproto.uint32_field(2)
    signature: "Signature" = betterproto.message_field(3)


@dataclass
class Signature(betterproto.Message):
    """Results of preparing a statement"""

    columns: List["ColumnMetaData"] = betterproto.message_field(1)
    sql: str = betterproto.string_field(2)
    parameters: List["AvaticaParameter"] = betterproto.message_field(3)
    cursor_factory: "CursorFactory" = betterproto.message_field(4)
    statement_type: "StatementType" = betterproto.enum_field(5)


@dataclass
class ColumnMetaData(betterproto.Message):
    ordinal: int = betterproto.uint32_field(1)
    auto_increment: bool = betterproto.bool_field(2)
    case_sensitive: bool = betterproto.bool_field(3)
    searchable: bool = betterproto.bool_field(4)
    currency: bool = betterproto.bool_field(5)
    nullable: int = betterproto.uint32_field(6)
    signed: bool = betterproto.bool_field(7)
    display_size: int = betterproto.uint32_field(8)
    label: str = betterproto.string_field(9)
    column_name: str = betterproto.string_field(10)
    schema_name: str = betterproto.string_field(11)
    precision: int = betterproto.uint32_field(12)
    scale: int = betterproto.uint32_field(13)
    table_name: str = betterproto.string_field(14)
    catalog_name: str = betterproto.string_field(15)
    read_only: bool = betterproto.bool_field(16)
    writable: bool = betterproto.bool_field(17)
    definitely_writable: bool = betterproto.bool_field(18)
    column_class_name: str = betterproto.string_field(19)
    type: "AvaticaType" = betterproto.message_field(20)


@dataclass(eq=False, repr=False)
class AvaticaType(betterproto.Message):
    """Base class for a column type"""

    id: int = betterproto.uint32_field(1)
    name: str = betterproto.string_field(2)
    rep: "Rep" = betterproto.enum_field(3)
    columns: List["ColumnMetaData"] = betterproto.message_field(4)
    component: "AvaticaType" = betterproto.message_field(5)


@dataclass
class AvaticaParameter(betterproto.Message):
    """Metadata for a parameter"""

    signed: bool = betterproto.bool_field(1)
    precision: int = betterproto.uint32_field(2)
    scale: int = betterproto.uint32_field(3)
    parameter_type: int = betterproto.uint32_field(4)
    type_name: str = betterproto.string_field(5)
    class_name: str = betterproto.string_field(6)
    name: str = betterproto.string_field(7)


@dataclass
class CursorFactory(betterproto.Message):
    """Information necessary to convert an Iterable into a Calcite Cursor"""

    style: "CursorFactoryStyle" = betterproto.enum_field(1)
    class_name: str = betterproto.string_field(2)
    field_names: List[str] = betterproto.string_field(3)


@dataclass
class Frame(betterproto.Message):
    """A collection of rows"""

    offset: int = betterproto.uint64_field(1)
    done: bool = betterproto.bool_field(2)
    rows: List["Row"] = betterproto.message_field(3)


@dataclass
class Row(betterproto.Message):
    """A row is a collection of values"""

    value: List["ColumnValue"] = betterproto.message_field(1)


@dataclass
class DatabaseProperty(betterproto.Message):
    """
    Database property, list of functions the database provides for a certain
    operation
    """

    name: str = betterproto.string_field(1)
    functions: List[str] = betterproto.string_field(2)


@dataclass
class WireMessage(betterproto.Message):
    """
    Message which encapsulates another message to support a single RPC endpoint
    """

    name: str = betterproto.string_field(1)
    wrapped_message: bytes = betterproto.bytes_field(2)


@dataclass
class ColumnValue(betterproto.Message):
    """A value might be a TypedValue or an Array of TypedValue's"""

    value: List["TypedValue"] = betterproto.message_field(1)
    array_value: List["TypedValue"] = betterproto.message_field(2)
    has_array_value: bool = betterproto.bool_field(3)
    scalar_value: "TypedValue" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class TypedValue(betterproto.Message):
    """
    Generic wrapper to support any SQL type. Struct-like to work around no
    polymorphism construct.
    """

    type: "Rep" = betterproto.enum_field(1)
    bool_value: bool = betterproto.bool_field(2)
    string_value: str = betterproto.string_field(3)
    number_value: int = betterproto.sint64_field(4)
    # includes numeric types and date/time types.
    bytes_value: bytes = betterproto.bytes_field(5)
    double_value: float = betterproto.double_field(6)
    null: bool = betterproto.bool_field(7)
    array_value: List["TypedValue"] = betterproto.message_field(8)
    component_type: "Rep" = betterproto.enum_field(9)
    implicitly_null: bool = betterproto.bool_field(10)


@dataclass
class MetaDataOperationArgument(betterproto.Message):
    """Represents the breadth of arguments to DatabaseMetaData functions"""

    string_value: str = betterproto.string_field(1)
    bool_value: bool = betterproto.bool_field(2)
    int_value: int = betterproto.sint32_field(3)
    string_array_values: List[str] = betterproto.string_field(4)
    int_array_values: List[int] = betterproto.sint32_field(5)
    type: "MetaDataOperationArgumentArgumentType" = betterproto.enum_field(6)


@dataclass
class QueryState(betterproto.Message):
    type: "StateType" = betterproto.enum_field(1)
    sql: str = betterproto.string_field(2)
    op: "MetaDataOperation" = betterproto.enum_field(3)
    args: List["MetaDataOperationArgument"] = betterproto.message_field(4)
    has_args: bool = betterproto.bool_field(5)
    has_sql: bool = betterproto.bool_field(6)
    has_op: bool = betterproto.bool_field(7)



import os
from setuptools import setup, find_packages

files = []
for item in os.walk("pyi/aiophoenixdb"):
    base_path = item[0][4:]
    files.append((base_path, [os.path.join(item[0], f) for f in item[-1]]))

with open("README.rst", "r") as f:
    long_desc_text = f.read()

setup(
    name="aiophoenixdb"
    , version="0.0.4"
    , author="Nick Hao"
    , author_email="nickwike72@gmail.com"
    , package_dir={"aiophoenixdb": "./src/aiophoenixdb"}
    , description="Asyncio with phoenixdb"
    , packages=find_packages(include=['aiophoenixdb*'], where="./src")
    , data_files=files
    , long_description_content_type="text/x-rst"
    , long_description=long_desc_text
    , install_requires=[
        "betterproto>=2.0.0b6",
        "aiohttp>=3.8.6",
        "requests-gssapi",
        "requests~=2.31.0",
        "gssapi"
    ]
    , classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cursors import Cursor, CursorRef
from .avatica.client import AvaticaClient
from typing import Generic, TypeVar, overload, Dict, Any, List
from aiophoenixdb.typeshed import Props, Self
from .meta import Meta

_C = TypeVar("_C", bound=Cursor)
_C2 = TypeVar("_C2", bound=Cursor)


class Connection(Generic[_C]):

    _client: AvaticaClient
    _closed: bool
    _cursors: List[CursorRef]
    cursor_factory: type[Cursor]
    _phoenix_props: Dict[str, Any]
    avatica_props_init: Dict[str, Any]
    _conn_id: str
    _avatica_props: Dict

    def __init__(self,
                 client: AvaticaClient,
                 cursor_factory: _C,
                 **kwargs
                 ): ...

    async def connect(self): ...
    async def __aenter__(self: Self) -> Self: ...
    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None: ...

    @property
    def client(self) -> AvaticaClient: ...
    @property
    def closed(self) -> bool: ...
    @property
    def connect_id(self) -> str: ...
    @property
    def _default_avatica_props(self): ...
    @staticmethod
    def _map_conn_props(conn_props: Props): ...
    @staticmethod
    def _map_legacy_avatica_props(props: Props): ...
    async def open(self) -> None: ...
    async def close(self) -> None: ...
    async def commit(self) -> None: ...
    async def rollback(self) -> None: ...

    def cursor(self) -> _C: ...
    @overload
    def cursor(self, cursor_factory: type[_C2] | None = ...) -> _C2: ...
    async def set_session(self, **props) -> None: ...

    @overload
    def autocommit(self) -> bool: ...
    @overload
    def autocommit(self, value) -> None: ...

    @overload
    def readonly(self) -> bool: ...
    @overload
    def readonly(self, value) -> None: ...

    @overload
    def transactionisolation(self) -> bool: ...
    @overload
    def transactionisolation(self, value) -> None: ...

    def meta(self) -> Meta: ...


# Copyright 2024 Nick Hao
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from _weakref import ReferenceType
from typing import Any, List, Generator, Dict, TypeVar, Callable
from aiophoenixdb.typeshed import Self
from aiophoenixdb.avatica.proto.common_pb import ColumnMetaData, Signature, Frame, Row
from aiophoenixdb.avatica.proto.responses_pb import ResultSetResponse, SyncResultsResponse
from aiophoenixdb.connection import Connection

_C = TypeVar("_C", bound="Cursor")

class CursorRef:

    _ref: ReferenceType

    def __init__(self, o: _C, callback: Callable[[CursorRef], _C] | None = ...): ...

    def __call__(self, *args, **kwargs) -> _C: ...

class Cursor(object):
    _ARRAY_SIZE: int
    _ITER_SIZE: int
    _connection: Connection
    _id: int
    _signature: Any
    _column_data_types: List
    _frame: Frame | None
    _pos: int
    _closed: bool
    _array_size: int
    _iter_size: int
    _update_count: int
    _parameter_data_types: List[Any]



    def __init__(self, connection: Connection, _id: int = None): ...
    async def __aenter__(self: Self) -> Self: ...
    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None: ...
    def __iter__(self: Self): Self
    async def __anext__(self) -> Generator[List[List[Any], Any]]: ...

    async def close(self) -> None: ...
    @property
    def closed(self) -> bool: ...

    @property
    def description(self) -> List[Any]: ...

    @staticmethod
    def _get_column_name(column: ColumnMetaData) -> str: ...

    async def _set_id(self, _id) -> None: ...

    def _set_signature(self, signature: Signature) -> None: ...
    def _set_frame(self, frame: Frame | None) -> None: ...

    async def _fetch_next_frame(self) -> None: ...

    async def process_result(self, result: ResultSetResponse) -> None: ...

    async def _process_results(self, results: List[ResultSetResponse]) -> None: ...

    def _transform_parameters(self, parameters) -> List[Any]: ...

    async def execute(self, operation, parameters=None) -> None: ...

    async def executemany(self, operation, seq_of_parameters) -> List[int]: ...

    async def get_sync_results(self, state) -> SyncResultsResponse: ...

    async def fetch(self, signature) -> None: ...

    def transform_row(self, row: Row) -> List[Any]: ...

    async def fetchone(self) -> Any: ...

    async def fetchmany(self, size=None) -> List[Any]: ...

    async def fetchall(self) -> List[Any]: ...

    def setinputsizes(self, sizes) -> None: ...

    def setoutputsize(self, size, column=None) -> None: ...

    @property
    def connection(self) -> Connection: ...

    @property
    def rowcount(self) -> int: ...

    @property
    def rownumber(self) -> int: ...

    def ref(self,  callback: Callable[[CursorRef], Any] | None = ...) -> CursorRef: ...


class DictCursor(Cursor):
    """A cursor which returns results as a dictionary"""

    def _transform_row(self, row) -> Dict[str, Any]: ...